#//----------------------------------------------------------------------
#//   Copyright 2011 Cypress Semiconductor
#//   Copyright 2010 Mentor Graphics Corporation
#//   Copyright 2011 Cadence Design Systems, Inc.
#//   Copyright 2019 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------
"""
Title: Resources

Topic: Intro

A resource is a parameterized container that holds arbitrary data.
Resources can be used to configure components, supply data to
sequences, or enable sharing of information across disparate parts of
a testbench.  They are stored using scoping information so their
visibility can be constrained to certain parts of the testbench.
Resource containers can hold any type of data, constrained only by
the data types available in SystemVerilog.  Resources can contain
scalar objects, class handles, queues, lists, or even virtual
interfaces.

Resources are stored in a resource database so that each resource can
be retrieved by name or by type. The database has both a name table
and a type table and each resource is entered into both. The database
is globally accessible.

Each resource has a set of scopes over which it is visible.  The set
of scopes is represented as a regular expression.  When a resource is
looked up the scope of the entity doing the looking up is supplied to
the lookup function.  This is called the ~current scope~.  If the
current scope is in the set of scopes over which a resource is
visible then the resource can be retuned in the lookup.

Resources can be looked up by name or by type. To support type lookup
each resource has a static type handle that uniquely identifies the
type of each specialized resource container.

Multiple resources that have the same name are stored in a queue.
Each resource is pushed into a queue with the first one at the front
of the queue and each subsequent one behind it.  The same happens for
multiple resources that have the same type.  The resource queues are
searched front to back, so those placed earlier in the queue have
precedence over those placed later.

The precedence of resources with the same name or same type can be
altered.  One way is to set the ~precedence~ member of the resource
container to any arbitrary value.  The search algorithm will return
the resource with the highest precedence.  In the case where there
are multiple resources that match the search criteria and have the
same (highest) precedence, the earliest one located in the queue will
be one returned.  Another way to change the precedence is to use the
set_priority function to move a resource to either the front or back
of the queue.

The classes defined here form the low level layer of the resource
database.  The classes include the resource container and the database
that holds the containers.  The following set of classes are defined
here:

<uvm_resource_types>: A class without methods or members, only
typedefs and enums. These types and enums are used throughout the
resources facility.  Putting the types in a class keeps them confined
to a specific name space.

<uvm_resource_options>: policy class for setting options, such
as auditing, which effect resources.

<uvm_resource_base>: the base (untyped) resource class living in the
resource database.  This class includes the interface for setting a
resource as read-only, notification, scope management, altering
search priority, and managing auditing.

<uvm_resource#(T)>: parameterized resource container.  This class
includes the interfaces for reading and writing each resource.
Because the class is parameterized, all the access functions are type
safe.

<uvm_resource_pool>: the resource database. This is a singleton
class object.
"""

from cocotb.triggers import Event

from .uvm_object import UVMObject
from .sv import sv, uvm_re_match, uvm_glob_to_re, wait
from ..uvm_macros import uvm_typename
from ..macros import uvm_info
from .uvm_globals import (uvm_report_error, uvm_report_warning)
from .uvm_object_globals import UVM_NONE
from .uvm_spell_chkr import UVMSpellChkr
from .uvm_pool import UVMPool
from .uvm_printer import UVMLinePrinter
from .uvm_queue import UVMQueue


#----------------------------------------------------------------------
# Class: uvm_resource_types
#
# Provides typedefs and enums used throughout the resources facility.
# This class has no members or methods, only typedefs.  It's used in
# lieu of package-scope types.  When needed, other classes can use
# these types by prefixing their usage with uvm_resource_types::.  E.g.
#
#|  uvm_resource_types::rsrc_q_t queue
#
#----------------------------------------------------------------------
#class uvm_resource_types
#
#  // types uses for setting overrides
#  typedef bit[1:0] override_t
#  typedef enum override_t { TYPE_OVERRIDE = 2'b01,
#                            NAME_OVERRIDE = 2'b10 } override_e
#
#   // general purpose queue of resourcex
#  typedef uvm_queue#(uvm_resource_base) rsrc_q_t
#
#  // enum for setting resource search priority
#  typedef enum { PRI_HIGH, PRI_LOW } priority_e
#
#  // access record for resources.  A set of these is stored for each
#  // resource by accessing object.  It's updated for each read/write.
#  typedef struct
#  {
#    time read_time
#    time write_time
#    int unsigned read_count
#    int unsigned write_count
#  } access_t
#
#endclass

TYPE_OVERRIDE = 1
NAME_OVERRIDE = 2

UVM_USE_RESOURCE_CONVERTER = False

PRI_HIGH = 0
PRI_LOW = 1


class Access_t:
    def __init__(self):
        self.read_time = 0
        self.write_time = 0
        self.read_count = 0
        self.write_count = 0


class UVMResourceOptions:
    """
    Provides a namespace for managing options for the
    resources facility.  The only thing allowed in this class is static
    local data members and static functions for manipulating and
    retrieving the value of the data members.  The static local data
    members represent options and settings that control the behavior of
    the resources facility.

    Options include:

     * auditing:  on/off

       The default for auditing is on.  You may wish to turn it off to
       for performance reasons.  With auditing off memory is not
       consumed for storage of auditing information and time is not
       spent collecting and storing auditing information.  Of course,
       during the period when auditing is off no audit trail information
       is available
    """


    auditing = True

    @classmethod
    def turn_on_auditing(cls):
        UVMResourceOptions.auditing = True

    @classmethod
    def turn_off_auditing(cls):
        UVMResourceOptions.auditing = False

    @classmethod
    def is_auditing(cls):
        return UVMResourceOptions.auditing


class get_t:
    def __init(self):
        self.name = ""
        self.scope = ""
        self.rsrc = None
        self.t = 0

#----------------------------------------------------------------------
# Class: uvm_resource_base
#
# Non-parameterized base class for resources.  Supports interfaces for
# scope matching, and virtual functions for printing the resource and
# for printing the accessor list
#----------------------------------------------------------------------


class UVMResourceBase(UVMObject):

    #// variable: default_precedence
    #//
    #// The default precedence for an resource that has been created.
    #// When two resources have the same precedence, the first resource
    #// found has precedence.
    #//
    default_precedence = 1000

    def __init__(self, name="", s="*"):
        """
        Function: new

        constructor for uvm_resource_base.  The constructor takes two
        arguments, the name of the resource and a regular expression which
        represents the set of scopes over which this resource is visible.
        Args:
            name:
            s:
        Raises:
        """
        UVMObject.__init__(self, name)
        if s is None:
            raise TypeError('s/scope must be string (or use "*" default) '
                    + ', got: ' + str(s))
        #self.scope = ""
        self.set_scope(s)
        self.modified = False
        self.read_only = False
        self.access = {}  # uvm_resource_types::access_t access[string]
        # variable: precedence
        #
        # This variable is used to associate a precedence that a resource
        # has with respect to other resources which match the same scope
        # and name. Resources are set to the <default_precedence> initially,
        # and may be set to a higher or lower precedence as desired.
        self.precedence = UVMResourceBase.default_precedence
        self.event_modified = Event(name + '_' + 'event_modified')
        self.modified = False

    def get_type_handle(self):
        """
        Function: get_type_handle

        Pure virtual function that returns the type handle of the resource
        container.
        Raises:
        """
        raise Exception('get_type_handle: pure virtual method')

    #//---------------------------
    #// Group: Read-only Interface
    #//---------------------------

    #// Function: set_read_only
    #//
    #// Establishes this resource as a read-only resource.  An attempt
    #// to call <uvm_resource#(T)::write> on the resource will cause an error.

    def set_read_only(self):
        self.read_only = True

    #// function set_read_write
    #//
    #// Returns the resource to normal read-write capability.
    #
    #// Implementation question: Not sure if this function is necessary.
    #// Once a resource is set to read_only no one should be able to change
    #// that.  If anyone can flip the read_only bit then the resource is not
    #// truly read_only.

    def set_read_write(self):
        self.read_only = False

    def is_read_only(self):
        """
        Function: is_read_only

        Returns one if this resource has been set to read-only, zero
        otherwise
        Returns:
        """
        return self.read_only

    #//--------------------
    #// Group: Notification
    #//--------------------

    #// Task: wait_modified
    #//
    #// This task blocks until the resource has been modified -- that is, a
    #// <uvm_resource#(T)::write> operation has been performed.  When a
    #// <uvm_resource#(T)::write> is performed the modified bit is set which
    #// releases the block.  Wait_modified() then clears the modified bit so
    #// it can be called repeatedly.


    async def wait_modified(self):
        """
        """
        await wait(lambda: self.modified is True, self.event_modified)
        self.modified = False

    #//-----------------------
    #// Group: Scope Interface
    #//-----------------------
    #//
    #// Each resource has a name, a value and a set of scopes over which it
    #// is visible. A scope is a hierarchical entity or a context.  A scope
    #// name is a multi-element string that identifies a scope.  Each
    #// element refers to a scope context and the elements are separated by
    #// dots (.).
    #//
    #//|    top.env.agent.monitor
    #//
    #// Consider the example above of a scope name.  It consists of four
    #// elements: "top", "env", "agent", and "monitor".  The elements are
    #// strung together with a dot separating each element.  ~top.env.agent~
    #// is the parent of ~top.env.agent.monitor~, ~top.env~ is the parent of
    #// ~top.env.agent~, and so on.  A set of scopes can be represented by a
    #// set of scope name strings.  A very straightforward way to represent
    #// a set of strings is to use regular expressions.  A regular
    #// expression is a special string that contains placeholders which can
    #// be substituted in various ways to generate or recognize a
    #// particular set of strings.  Here are a few simple examples:
    #//
    #//|     top\..*	                all of the scopes whose top-level component
    #//|                            is top
    #//|    top\.env\..*\.monitor	all of the scopes in env that end in monitor
    #//|                            i.e. all the monitors two levels down from env
    #//|    .*\.monitor	            all of the scopes that end in monitor; i.e.
    #//|                            all the monitors (assuming a naming convention
    #//|                            was used where all monitors are named "monitor")
    #//|    top\.u[1-5]\.*	        all of the scopes rooted and named u1, u2, u3,
    #//                             u4, or u5, and any of their subscopes.
    #//
    #// The examples above use POSIX regular expression notation.  This is
    #// a very general and expressive notation.  It is not always the case
    #// that so much expressiveness is required.  Sometimes an expression
    #// syntax that is easy to read and easy to write is useful, even if
    #// the syntax is not as expressive as the full power of POSIX regular
    #// expressions.  A popular substitute for regular expressions is
    #// globs.  A glob is a simplified regular expression. It only has
    #// three metacharacters -- *, +, and ?.  Character ranges are not
    #// allowed and dots are not a metacharacter in globs as they are in
    #// regular expressions.  The following table shows glob
    #// metacharacters.
    #//
    #//|      char	meaning	                regular expression
    #//|                                    equivalent
    #//|      *	    0 or more characters	.*
    #//|      +	    1 or more characters	.+
    #//|      ?	    exactly one character	.
    #//
    #// Of the examples above, the first three can easily be translated
    #// into globs.  The last one cannot.  It relies on notation that is
    #// not available in glob syntax.
    #//
    #//|    regular expression	    glob equivalent
    #//|    ---------------------      ------------------
    #//|    top\..*	            top.*
    #//|    top\.env\..*\.monitor	    top.env.*.monitor
    #//|    .*\.monitor	            *.monitor
    #//
    #// The resource facility supports both regular expression and glob
    #// syntax.  Regular expressions are identified as such when they
    #// surrounded by '/' characters. For example, ~/^top\.*/~ is
    #// interpreted as the regular expression ~^top\.*~, where the
    #// surrounding '/' characters have been removed. All other expressions
    #// are treated as glob expressions. They are converted from glob
    #// notation to regular expression notation internally.  Regular expression
    #// compilation and matching as well as glob-to-regular expression
    #// conversion are handled by two DPI functions:
    #//
    #//|    function int uvm_re_match(string re, string str)
    #//|    function string uvm_glob_to_re(string glob)
    #//
    #// uvm_re_match both compiles and matches the regular expression.
    #// All of the matching is done using regular expressions, so globs are
    #// converted to regular expressions and then processed.


    def set_scope(self, s):
        """
        Function: set_scope

        Set the value of the regular expression that identifies the set of
        scopes over which this resource is visible.  If the supplied
        argument is a glob it will be converted to a regular expression
        before it is stored.

        Args:
            s:
        """
        scope = uvm_glob_to_re(s)
        self.scope = scope

    def get_scope(self):
        """
        Function: get_scope

        Retrieve the regular expression string that identifies the set of
        scopes over which this resource is visible.

        Returns:
        """
        return self.scope

    def match_scope(self, s):
        """
        Function: match_scope

        Using the regular expression facility, determine if this resource
        is visible in a scope.  Return one if it is, zero otherwise.

        Args:
            s:
        Returns:
        """
        err = uvm_re_match(self.scope, s)
        return (err == 0)

    #//----------------
    #// Group: Priority
    #//----------------
    #//
    #// Functions for manipulating the search priority of resources.  The
    #// function definitions here are pure virtual and are implemented in
    #// derived classes.  The definitons serve as a priority management
    #// interface.

    #// Function: set priority
    #//
    #// Change the search priority of the resource based on the value of
    #// the priority enum argument.
    #//
    #pure virtual function void set_priority (uvm_resource_types::priority_e pri)

    #//-------------------------
    #// Group: Utility Functions
    #//-------------------------

    #// function convert2string
    #//
    #// Create a string representation of the resource value.  By default
    #// we don't know how to do this so we just return a "?".  Resource
    #// specializations are expected to override this function to produce a
    #// proper string representation of the resource value.

    #function string convert2string()
    #  return "?"
    #endfunction

    #// Function: do_print
    #//
    #// Implementation of do_print which is called by print().

    #function void do_print (uvm_printer printer)
    #  printer.print_string("",$sformatf("%s [%s] : %s", get_name(), get_scope(), convert2string()))
    #endfunction


    def record_read_access(self, accessor=None):
        """
        #-------------------
        Group: Audit Trail
        #-------------------

        To find out what is happening as the simulation proceeds, an audit
        trail of each read and write is kept. The <uvm_resource#(T)::read> and
        <uvm_resource#(T)::write> methods each take an accessor argument.  This is a
        handle to the object that performed that resource access.

        .. code-block:: python

            function T read(uvm_object accessor = None)
            function void write(T t, uvm_object accessor = None)

        The accessor can by anything as long as it is derived from
        uvm_object.  The accessor object can be a component or a sequence
        or whatever object from which a read or write was invoked.
        Typically the `this` handle is used as the
        accessor.  For example:

        .. code-block:: python

            uvm_resource#(int) rint
            int i
            ...
            rint.write(7, this)
            i = rint.read(this)

        The accessor's ~get_full_name()~ is stored as part of the audit trail.
        This way you can find out what object performed each resource access.
        Each audit record also includes the time of the access (simulation time)
        and the particular operation performed (read or write).

        Auditing is controlled through the `uvm_resource_options` class.
        function: record_read_access
        #function void record_read_access(uvm_object accessor = None)
        Args:
            accessor:
        """
        _str = ""
        access_record = None  # uvm_resource_types::access_t access_record
        #  // If an accessor object is supplied then get the accessor record.
        #  // Otherwise create a new access record.  In either case populate
        #  // the access record with information about this access.  Check
        #  // first to make sure that auditing is turned on.
        if not UVMResourceOptions.is_auditing():
            return

        # If an accessor is supplied, then use its name
        # as the database entry for the accessor record.
        #  Otherwise, use "<empty>" as the database entry.
        if accessor is not None:
            _str = accessor.get_full_name()
        else:
            _str = "<empty>"

        # Create a new accessor record if one does not exist
        if _str in self.access:
            access_record = self.access[_str]
        else:
            access_record = Access_t()
            self.init_access_record(access_record)

        # Update the accessor record
        access_record.read_count += 1
        # TODO access_record.read_time = $realtime
        self.access[_str] = access_record

    def record_write_access(self, accessor=None):
        """
        function: record_write_access

        Args:
            accessor:
        """
        # If an accessor object is supplied then get the accessor record.
        # Otherwise create a new access record.  In either case populate
        # the access record with information about this access.  Check
        # first that auditing is turned on
        if UVMResourceOptions.is_auditing():
            if accessor is not None:
                access_record = None  # uvm_resource_types::access_t
                _str = accessor.get_full_name()
                if _str in self.access:
                    access_record = self.access[_str]
                else:
                    access_record = Access_t()
                    self.init_access_record(access_record)
                access_record.write_count += 1
                # TODO access_record.write_time = $realtime
                self.access[_str] = access_record


    #// Function: print_accessors
    #//
    #// Dump the access records for this resource
    #//
    #virtual function void print_accessors()

    #  string str
    #  uvm_component comp
    #  uvm_resource_types::access_t access_record
    #  string qs[$]
    #
    #  if(access.num() == 0)
    #    return

    #  foreach (access[i]):
    #    str = i
    #    access_record = access[str]
    #    qs.push_back($sformatf("%s reads: %0d @ %0t  writes: %0d @ %0t\n",str,
    #             access_record.read_count,
    #             access_record.read_time,
    #             access_record.write_count,
    #             access_record.write_time))
    #  end
    #  `uvm_info("UVM/RESOURCE/ACCESSOR",`UVM_STRING_QUEUE_STREAMING_PACK(qs),UVM_NONE)

    #endfunction

    def init_access_record(self, access_record):
        """
        Function: init_access_record

        Initialize a new access record

        Args:
            access_record:
        """
        # inout uvm_resource_types::access_t access_record)
        access_record.read_time = 0
        access_record.write_time = 0
        access_record.read_count = 0
        access_record.write_count = 0

#//----------------------------------------------------------------------
#// Class: uvm_resource #(T)
#//
#// Parameterized resource.  Provides essential access methods to read
#// from and write to the resource database.
#//----------------------------------------------------------------------

class UVMResource(UVMResourceBase):

    my_type = None

    def __init__(self, name="", scope=""):
        UVMResourceBase.__init__(self, name, scope)
        self.val = None   # Resource value
        self.m_r2s = None  # Res2string converter

    def convert2string(self):
        if UVM_USE_RESOURCE_CONVERTER:
            self.m_get_converter()
            return self.m_r2s.convert2string(self.val)
        return sv.sformatf("(%s) %s", uvm_typename(self.val), self.val)

    #//----------------------
    #// Group: Type Interface
    #//----------------------
    #//
    #// Resources can be identified by type using a static type handle.
    #// The parent class provides the virtual function interface
    #// <get_type_handle>.  Here we implement it by returning the static type
    #// handle.

    #// Function: get_type
    #//
    #// Static function that returns the static type handle.  The return
    #// type is this_type, which is the type of the parameterized class.

    @classmethod
    def get_type(cls):
        if UVMResource.my_type is None:
            UVMResource.my_type = UVMResource()
        else:
            return UVMResource.my_type

    def get_type_handle(self):
        """
        Function: get_type_handle

        Returns the static type handle of this resource in a polymorphic
        fashion.  The return type of get_type_handle() is
        uvm_resource_base.  This function is not static and therefore can
        only be used by instances of a parameterized resource.
        Returns:
        """
        return UVMResource.get_type()

    #//-------------------------
    #// Group: Set/Get Interface
    #//-------------------------
    #//
    #// uvm_resource#(T) provides an interface for setting and getting a
    #// resources.  Specifically, a resource can insert itself into the
    #// resource pool.  It doesn't make sense for a resource to get itself,
    #// since you can't call a function on a handle you don't have.
    #// However, a static get interface is provided as a convenience.  This
    #// obviates the need for the user to get a handle to the global
    #// resource pool as this is done for him here.

    def set(self):
        """
        Function: set

        Simply put this resource into the global resource pool
        """
        """ Simply put this resource into the global resource pool"""
        pool = UVMResourcePool.get()
        pool.set(self)

    def set_override(self, override=NAME_OVERRIDE):
        """
        Function: set_override

        Put a resource into the global resource pool as an override.  This
        means it gets put at the head of the list and is searched before
        other existing resources that occupy the same position in the name
        map or the type map.  The default is to override both the name and
        type maps.  However, using the `override` argument you can specify
        that either the name map or type map is overridden.
        Args:
            override:
            NAME_OVERRIDE:
        """
        rp = UVMResourcePool.get()
        rp.set(self, override)

    @classmethod
    def get_by_name(cls, scope, name, rpterr=True):
        """
        Function: get_by_name

        looks up a resource by `name` in the name map. The first resource
        with the specified name, whose type is the current type, and is
        visible in the specified `scope` is returned, if one exists.  The
        `rpterr` flag indicates whether or not an error should be reported
        if the search fails.  If `rpterr` is set to one then a failure
        message is issued, including suggested spelling alternatives, based
        on resource names that exist in the database, gathered by the spell
        checker.
        Args:
            cls:
            scope:
            name:
            rpterr:
        Returns:
        """
        rp = UVMResourcePool.get()
        #rsrc = None
        rsrc_base = rp.get_by_name(scope, name, cls.get_type(), rpterr)
        if rsrc_base is None:
            return None
        return rsrc_base

    #// Function: get_by_type
    #//
    #// looks up a resource by ~type_handle~ in the type map. The first resource
    #// with the specified ~type_handle~ that is visible in the specified ~scope~ is
    #// returned, if one exists. If there is no resource matching the specifications,
    #// ~None~ is returned.

    #static function this_type get_by_type(string scope = "",
    #                                      uvm_resource_base type_handle)

    #  uvm_resource_pool rp = uvm_resource_pool::get()
    #  uvm_resource_base rsrc_base
    #  this_type rsrc
    #  string msg

    #  if(type_handle is None)
    #    return None

    #  rsrc_base = rp.get_by_type(scope, type_handle)
    #  if(rsrc_base is None)
    #    return None

    #  if(!$cast(rsrc, rsrc_base)):
    #    $sformat(msg, "Resource with specified type handle in scope %s was not located", scope)
    #    `uvm_warning("RSRCNF", msg)
    #    return None
    #  end

    #  return rsrc

    #endfunction

    #//----------------------------
    #// Group: Read/Write Interface
    #//----------------------------
    #//
    #// <read> and <write> provide a type-safe interface for getting and
    #// setting the object in the resource container.  The interface is
    #// type safe because the value argument for <write> and the return
    #// value of <read> are T, the type supplied in the class parameter.
    #// If either of these functions is used in an incorrect type context
    #// the compiler will complain.

    #// Function: read
    #//
    #// Return the object stored in the resource container.  If an ~accessor~
    #// object is supplied then also update the accessor record for this
    #// resource.

    def read(self, accessor=None):
        """ Return the object stored in the resource container """
        self.record_read_access(accessor)
        return self.val

    def write(self, t, accessor=None):
        """
        Function: write
        Modify the object stored in this resource container.  If the
        resource is read-only then issue an error message and return
        without modifying the object in the container.  If the resource is
        not read-only and an `accessor` object has been supplied then also
        update the accessor record.  Lastly, replace the object value in
        the container with the value supplied as the argument, `t`, and
        release any processes blocked on
        <uvm_resource_base::wait_modified>.  If the value to be written is
        the same as the value already present in the resource then the
        write is not done.  That also means that the accessor record is not
        updated and the modified bit is not set.
        Args:
            t:
            accessor:
        """
        if self.is_read_only():
            uvm_report_error("resource", sv.sformatf("resource %s is read only -- cannot modify",
                self.get_name()))
            return

        # Set the modified bit and record the transaction only if the value
        # has actually changed.
        if self.val == t:
            return

        self.record_write_access(accessor)

        # set the value and set the dirty bit
        self.val = t
        self.modified = True
        self.event_modified.set()

    #//----------------
    #// Group: Priority
    #//----------------
    #//
    #// Functions for manipulating the search priority of resources.  These
    #// implementations of the interface defined in the base class delegate
    #// to the resource pool.

    #// Function: set priority
    #//
    #// Change the search priority of the resource based on the value of
    #// the priority enum argument, ~pri~.

    #function void set_priority (uvm_resource_types::priority_e pri)
    #  uvm_resource_pool rp = uvm_resource_pool::get()
    #  rp.set_priority(this, pri)
    #endfunction

    @classmethod
    def get_highest_precedence(cls, q, T=None):
        """
        Function: get_highest_precedence

        In a queue of resources, locate the first one with the highest
        precedence whose type is T.  This function is static so that it can
        be called from anywhere.
        #static function this_type get_highest_precedence(ref uvm_resource_types::rsrc_q_t q)
        Args:
            q:
            T:
        Returns:
        """
        rsrc = None
        r = None
        i = 0
        prec = 0
        first = 0

        if len(q) == 0:
            return None

        # Locate first resources in the queue whose type is T
        for i in range(len(q)):
            first = i
            q_val = q[first]
            if T is not None:
                arr_rsrc = []
                if sv.cast(arr_rsrc, q_val, T):
                    rsrc = q_val
                    break
            else:
                rsrc = q_val
                break

        # no resource in the queue whose type is T
        if rsrc is None:
            return None

        prec = rsrc.precedence

        # start searching from the next resource after the first resource
        # whose type is T
        # for(int i = first+1; i < q.size(); ++i):
        for i in range(first + 1, len(q)):
            if T is not None:
                arr_r = []
                if sv.cast(arr_r, q[i], T):
                    r = q[i]
                    if r.precedence > prec:
                        rsrc = r
                        prec = r.precedence
            else:
                r = q[i]
                if r.precedence > prec:
                    rsrc = r
                    prec = r.precedence
        return rsrc

#//----------------------------------------------------------------------
#// Class: uvm_resource_pool
#//
#// The global (singleton) resource database.
#//
#// Each resource is stored both by primary name and by type handle.  The
#// resource pool contains two associative arrays, one with name as the
#// key and one with the type handle as the key.  Each associative array
#// contains a queue of resources.  Each resource has a regular
#// expression that represents the set of scopes over which it is visible.
#//
#//|  +------+------------+                          +------------+------+
#//|  | name | rsrc queue |                          | rsrc queue | type |
#//|  +------+------------+                          +------------+------+
#//|  |      |            |                          |            |      |
#//|  +------+------------+                  +-+-+   +------------+------+
#//|  |      |            |                  | | |<--+---*        |  T   |
#//|  +------+------------+   +-+-+          +-+-+   +------------+------+
#//|  |  A   |        *---+-->| | |           |      |            |      |
#//|  +------+------------+   +-+-+           |      +------------+------+
#//|  |      |            |      |            |      |            |      |
#//|  +------+------------+      +-------+  +-+      +------------+------+
#//|  |      |            |              |  |        |            |      |
#//|  +------+------------+              |  |        +------------+------+
#//|  |      |            |              V  V        |            |      |
#//|  +------+------------+            +------+      +------------+------+
#//|  |      |            |            | rsrc |      |            |      |
#//|  +------+------------+            +------+      +------------+------+
#//
#// The above diagrams illustrates how a resource whose name is A and
#// type is T is stored in the pool.  The pool contains an entry in the
#// type map for type T and an entry in the name map for name A.  The
#// queues in each of the arrays each contain an entry for the resource A
#// whose type is T.  The name map can contain in its queue other
#// resources whose name is A which may or may not have the same type as
#// our resource A.  Similarly, the type map can contain in its queue
#// other resources whose type is T and whose name may or may not be A.
#//
#// Resources are added to the pool by calling <set>; they are retrieved
#// from the pool by calling <get_by_name> or <get_by_type>.  When an object
#// creates a new resource and calls <set> the resource is made available to be
#// retrieved by other objects outside of itself; an object gets a
#// resource when it wants to access a resource not currently available
#// in its scope.
#//
#// The scope is stored in the resource itself (not in the pool) so
#// whether you get by name or by type the resource's visibility is
#// the same.
#//
#// As an auditing capability, the pool contains a history of gets.  A
#// record of each get, whether by <get_by_type> or <get_by_name>, is stored
#// in the audit record.  Both successful and failed gets are recorded. At
#// the end of simulation, or any time for that matter, you can dump the
#// history list.  This will tell which resources were successfully
#// located and which were not.  You can use this information
#// to determine if there is some error in name, type, or
#// scope that has caused a resource to not be located or to be incorrectly
#// located (i.e. the wrong resource is located).
#//
#//----------------------------------------------------------------------


class UVMResourcePool:

    rp = None

    def __init__(self):
        self.rtab = UVMPool()
        self.ttab = {}
        self.get_record = []  # History of gets

    @classmethod
    def get(cls):
        """
        Function: get

        Returns the singleton handle to the resource pool
        Returns:
        """
        if UVMResourcePool.rp is None:
            UVMResourcePool.rp = UVMResourcePool()
        return UVMResourcePool.rp

    def spell_check(self, s):
        """
        Function: spell_check

        Invokes the spell checker for a string s.  The universe of
        correctly spelled strings -- i.e. the dictionary -- is the name
        map.
        Args:
            s:
        Returns:
        """
        return UVMSpellChkr.check(self.rtab, s)

    #-----------
    # Group: Set
    #-----------

    def set(self, rsrc, override=False):
        """
        Function: set

        Add a new resource to the resource pool.  The resource is inserted
        into both the name map and type map so it can be located by
        either.

        An object creates a resources and `sets` it into the resource pool.
        Later, other objects that want to access the resource must `get` it
        from the pool

        Overrides can be specified using this interface.  Either a name
        override, a type override or both can be specified.  If an
        override is specified then the resource is entered at the front of
        the queue instead of at the back.  It is not recommended that users
        specify the override parameter directly, rather they use the
        `set_override`, `set_name_override`, or `set_type_override`
        functions.

        Args:
            rsrc:
            override:
        """
        rq = list()
        name = ""
        type_handle = None
        # If resource handle is ~None~ then there is nothing to do.
        if rsrc is None:
            return
        # insert into the name map.  Resources with empty names are
        # anonymous resources and are not entered into the name map
        name = rsrc.get_name()
        if name != "":
            if name in self.rtab:
                rq = self.rtab[name]
            else:
                rq = list()

        # Insert the resource into the queue associated with its name.
        # If we are doing a name override then insert it in the front of
        # the queue, otherwise insert it in the back.
        if override & NAME_OVERRIDE:
            rq.insert(0, rsrc)
        else:
            rq.append(rsrc)

        self.rtab[name] = rq

        # insert into the type map
        type_handle = rsrc.get_type_handle()
        if type_handle in self.ttab:
            rq = self.ttab[type_handle]
        else:
            rq = list()

        # insert the resource into the queue associated with its type.  If
        # we are doing a type override then insert it in the front of the
        # queue, otherwise insert it in the back of the queue.
        if override & TYPE_OVERRIDE:
            rq.insert(0, rsrc)
        else:
            rq.append(rsrc)
        self.ttab[type_handle] = rq


    #// Function: set_override
    #//
    #// The resource provided as an argument will be entered into the pool
    #// and will override both by name and type.

    #function void set_override(uvm_resource_base rsrc)
    #  set(rsrc, (uvm_resource_types::NAME_OVERRIDE |
    #             uvm_resource_types::TYPE_OVERRIDE))
    #endfunction


    #// Function: set_name_override
    #//
    #// The resource provided as an argument will entered into the pool
    #// using normal precedence in the type map and will override the name.

    #function void set_name_override(uvm_resource_base rsrc)
    #  set(rsrc, uvm_resource_types::NAME_OVERRIDE)
    #endfunction


    #// Function: set_type_override
    #//
    #// The resource provided as an argument will be entered into the pool
    #// using normal precedence in the name map and will override the type.

    #function void set_type_override(uvm_resource_base rsrc)
    #  set(rsrc, uvm_resource_types::TYPE_OVERRIDE)
    #endfunction


    def push_get_record(self, name, scope, rsrc):
        """
        function - push_get_record

        Insert a new record into the get history list.
        #function void push_get_record(string name, string scope,
                                       uvm_resource_base rsrc)
        Args:
            name:
            scope:
            rsrc:
        """
        impt = None  # get_t impt

        #  // if auditing is turned off then there is no reason
        #  // to save a get record
        if not UVMResourceOptions.is_auditing():
            return

        impt = get_t()

        impt.name  = name
        impt.scope = scope
        impt.rsrc  = rsrc
        impt.t     = sv.realtime()

        self.get_record.append(impt)
        #endfunction

    #// function - dump_get_records
    #//
    #// Format and print the get history list.
    #function void dump_get_records()

        #  get_t record
        #  bit success
        #  string qs[$]

        #  qs.push_back("--- resource get records ---\n")
        #  foreach (get_record[i]):
        #    record = get_record[i]
        #    success = (record.rsrc != None)
        #    qs.push_back($sformatf("get: name=%s  scope=%s  %s @ %0t\n",
        #             record.name, record.scope,
        #             ((success)?"success":"fail"),
        #             record.t))
        #  end
        #  `uvm_info("UVM/RESOURCE/GETRECORD",`UVM_STRING_QUEUE_STREAMING_PACK(qs),UVM_NONE)
        #endfunction


    #--------------
    # Group: Lookup
    #--------------
    #
    # This group of functions is for finding resources in the resource database.
    #
    # <lookup_name> and <lookup_type> locate the set of resources that
    # matches the name or type (respectively) and is visible in the
    # current scope.  These functions return a queue of resources.
    #
    # <get_highest_precedence> traverse a queue of resources and
    # returns the one with the highest precedence -- i.e. the one whose
    # precedence member has the highest value.
    #
    # <get_by_name> and <get_by_type> use <lookup_name> and <lookup_type>
    # (respectively) and <get_highest_precedence> to find the resource with
    # the highest priority that matches the other search criteria.


    def lookup_name(self, scope, name, type_handle=None, rpterr=True):
        """
        Function: lookup_name

        Lookup resources by `name`.  Returns a queue of resources that
        match the `name`, `scope`, and `type_handle`.  If no resources
        match the queue is returned empty. If `rpterr` is set then a
        warning is issued if no matches are found, and the spell checker is
        invoked on `name`.  If `type_handle` is `None` then a type check is
        not made and resources are returned that match only `name` and
        `scope`.
        Args:
            scope:
            name:
            type_handle:
            rpterr:
        Returns:
        """
        rq = list()
        q = list()
        # rsrc = None
        r = None

        # ensure rand stability during lookup
        # process p = process::self()
        # string s
        # if(p!=None) s=p.get_randstate()
        # q=new()
        # if(p!=None) p.set_randstate(s)

        # resources with empty names are anonymous and do not exist in the name map
        if name == "":
            return q

        # Does an entry in the name map exist with the specified name?
        # If not, then we're done
        if name not in self.rtab:
            if rpterr:
                self.spell_check(name)
            return q

        rq = self.rtab[name]
        for i in range(0, len(rq)):
            r = rq[i]
            # does the type and scope match?
            handle_ok = type_handle is None or (r.get_type_handle() == type_handle)
            if handle_ok and r.match_scope(scope):
                q.append(r)

        return q

    def get_highest_precedence(self, q):
        """
        Function: get_highest_precedence

        Traverse a queue, `q`, of resources and return the one with the highest
        precedence.  In the case where there exists more than one resource
        with the highest precedence value, the first one that has that
        precedence will be the one that is returned.

        #function uvm_resource_base get_highest_precedence(ref uvm_resource_types::rsrc_q_t q)
        Args:
            q:
        Returns:
        """

        rsrc = None  # uvm_resource_base
        r = None  # uvm_resource_base
        i = 0
        prec = 0

        if (len(q) == 0):
            return None

        # get the first resources in the queue
        rsrc = q[0]
        prec = rsrc.precedence

        # start searching from the second resource
        for i in range(1, len(q)):
            r = q[i]
            if r.precedence > prec:
                rsrc = r
                prec = r.precedence
        return rsrc
        #
        #endfunction

    @classmethod
    def sort_by_precedence(cls, q):
        """
        Function: sort_by_precedence

        Given a list of resources, obtained for example from `lookup_scope`,
        sort the resources in  precedence order. The highest precedence
        resource will be first in the list and the lowest precedence will
        be last. Resources that have the same precedence and the same name
        will be ordered by most recently set first.
        #static function queue sort_by_precedence(ref uvm_resource_types::rsrc_q_t q)
        Args:
            cls:
            q:
        Returns:
        """
        #  uvm_resource_types::rsrc_q_t all[int]
        _all = {}
        r = None  # uvm_resource_base

        for i in range(len(q)):
            r = q[i]
            if r.precedence not in _all:
                _all[r.precedence] = UVMQueue()
            _all[r.precedence].push_front(r)  # since we will push_front in the final
        qq = []
        for key in _all:
            for j in range(len(_all[key])):
                r = _all[key].get(j)
                #qq.push_front(r)
                qq.insert(0, r)
        return qq
        #endfunction

    def get_by_name(self, scope, name, type_handle, rpterr=True):
        """
        Function: get_by_name

        Lookup a resource by `name`, `scope`, and `type_handle`.  Whether
        the get succeeds or fails, save a record of the get attempt.  The
        `rpterr` flag indicates whether to report errors or not.
        Essentially, it serves as a verbose flag.  If set then the spell
        checker will be invoked and warnings about multiple resources will
        be produced.

        #function uvm_resource_base get_by_name(string scope = "",
                                              string name,
                                              uvm_resource_base type_handle,
                                              bit rpterr = 1)
        Args:
            scope:
            name:
            type_handle:
            rpterr:
        Returns:
        """
        q = []  # uvm_resource_types::rsrc_q_t q
        rsrc = None  # uvm_resource_base rsrc
        q = self.lookup_name(scope, name, type_handle, rpterr)
        #
        if len(q) == 0:
            self.push_get_record(name, scope, None)
            return None

        rsrc = self.get_highest_precedence(q)
        self.push_get_record(name, scope, rsrc)
        return rsrc
        #endfunction

    #
    #// Function: lookup_type
    #//
    #// Lookup resources by type. Return a queue of resources that match
    #// the ~type_handle~ and ~scope~.  If no resources match then the returned
    #// queue is empty.
    #
    #function uvm_resource_types::rsrc_q_t lookup_type(string scope = "",
    #                                                  uvm_resource_base type_handle)
    #
    #  uvm_resource_types::rsrc_q_t q = new()
    #  uvm_resource_types::rsrc_q_t rq
    #  uvm_resource_base r
    #  int unsigned i
    #
    #  if(type_handle is None || !ttab.exists(type_handle)):
    #    return q
    #  end
    #
    #  rq = ttab[type_handle]
    #  for(int i = 0; i < rq.size(); ++i):
    #    r = rq.get(i)
    #    if(r.match_scope(scope))
    #      q.push_back(r)
    #  end
    #
    #  return q
    #
    #endfunction

    #// Function: get_by_type
    #//
    #// Lookup a resource by ~type_handle~ and ~scope~.  Insert a record into
    #// the get history list whether or not the get succeeded.
    #
    #function uvm_resource_base get_by_type(string scope = "",
    #                                       uvm_resource_base type_handle)
    #
    #  uvm_resource_types::rsrc_q_t q
    #  uvm_resource_base rsrc
    #
    #  q = lookup_type(scope, type_handle)
    #
    #  if(q.size() == 0):
    #    push_get_record("<type>", scope, None)
    #    return None
    #  end
    #
    #  rsrc = q.get(0)
    #  push_get_record("<type>", scope, rsrc)
    #  return rsrc
    #
    #endfunction


    def lookup_regex_names(self, scope, name, type_handle=None):
        """
        Function: lookup_regex_names

        This utility function answers the question, for a given `name`,
        `scope`, and `type_handle`, what are all of the resources with requested name,
        a matching scope (where the resource scope may be a
        regular expression), and a matching type?
        `name` and `scope` are explicit values.
        #function uvm_resource_types::rsrc_q_t lookup_regex_names(string scope,
                                                                string name,
                                                                uvm_resource_base type_handle = None)
        Args:
            scope:
            name:
            type_handle:
        Returns:
        """
        return self.lookup_name(scope, name, type_handle, 0)

    #// Function: lookup_regex
    #//
    #// Looks for all the resources whose name matches the regular
    #// expression argument and whose scope matches the current scope.
    #function uvm_resource_types::rsrc_q_t lookup_regex(string re, scope)
    #
    #  uvm_resource_types::rsrc_q_t rq
    #  uvm_resource_types::rsrc_q_t result_q
    #  int unsigned i
    #  uvm_resource_base r
    #
    #  re = uvm_glob_to_re(re)
    #  result_q = new()
    #
    #  foreach (rtab[name]):
    #    if(uvm_re_match(re, name))
    #      continue
    #    rq = rtab[name]
    #    for(i = 0; i < rq.size(); i++):
    #      r = rq.get(i)
    #      if(r.match_scope(scope))
    #        result_q.push_back(r)
    #    end
    #  end
    #
    #  return result_q
    #endfunction

    def lookup_scope(self, scope):
        """
        Function: lookup_scope

        This is a utility function that answers the question: For a given
        `scope`, what resources are visible to it?  Locate all the resources
        that are visible to a particular scope.  This operation could be
        quite expensive, as it has to traverse all of the resources in the
        database.
        #function uvm_resource_types::rsrc_q_t lookup_scope(string scope)
        Args:
            scope:
        Returns:
        """
        rq = None  # uvm_resource_types::rsrc_q_t rq
        r = None  # uvm_resource_base r
        i = 0
        q = UVMQueue()  # uvm_resource_types::rsrc_q_t q = new()
        # iterate in reverse order for the special case of autoconfig
        # of arrays. The array name with no [] needs to be higher priority.
        # This has no effect an manual accesses.
        name = ""
        if self.rtab.has_last():
            name = self.rtab.last()
            while True:
                rq = self.rtab[name]
                for i in range(len(rq)):
                    r = rq[i]
                    if r.match_scope(scope):
                        q.push_back(r)
                if self.rtab.has_prev():
                    name = self.rtab.prev()
                else:
                    break
        return q

    #//--------------------
    #// Group: Set Priority
    #//--------------------
    #//
    #// Functions for altering the search priority of resources.  Resources
    #// are stored in queues in the type and name maps.  When retrieving
    #// resources, either by type or by name, the resource queue is search
    #// from front to back.  The first one that matches the search criteria
    #// is the one that is returned.  The ~set_priority~ functions let you
    #// change the order in which resources are searched.  For any
    #// particular resource, you can set its priority to UVM_HIGH, in which
    #// case the resource is moved to the front of the queue, or to UVM_LOW in
    #// which case the resource is moved to the back of the queue.

    #// function- set_priority_queue
    #//
    #// This function handles the mechanics of moving a resource to either
    #// the front or back of the queue.

    def set_priority_queue(self, rsrc, q, pri):
        """
        #local function void set_priority_queue(uvm_resource_base rsrc,
                                              ref uvm_resource_types::rsrc_q_t q,
                                              uvm_resource_types::priority_e pri)
        Args:
            rsrc:
            q:
            pri:
        """
        r = None  # uvm_resource_base
        i = 0

        msg = ""
        name = rsrc.get_name()

        for i in range(len(q)):
            r = q[i]
            if (r == rsrc):
                break

        if r != rsrc:
            msg = sv.sformatf("Handle for res named %s not in the name name; cannot change its priority",
                name)
            uvm_report_error("NORSRC", msg)
            return

        q.pop(i)

        if pri == PRI_HIGH:
            q.append(rsrc)
        elif pri == PRI_LOW:
            q.append(rsrc)

        #endfunction


    #// Function: set_priority_type
    #//
    #// Change the priority of the ~rsrc~ based on the value of ~pri~, the
    #// priority enum argument.  This function changes the priority only in
    #// the type map, leaving the name map untouched.

    #function void set_priority_type(uvm_resource_base rsrc,
    #                                uvm_resource_types::priority_e pri)

    #  uvm_resource_base type_handle
    #  string msg
    #  uvm_resource_types::rsrc_q_t q

    #  if(rsrc is None):
    #    uvm_report_warning("NoneRASRC", "attempting to change the serach priority of a None resource")
    #    return
    #  end

    #  type_handle = rsrc.get_type_handle()
    #  if(!ttab.exists(type_handle)):
    #    $sformat(msg, "Type handle for resrouce named %s not found in type map; cannot change its search priority", rsrc.get_name())
    #    uvm_report_error("RNFTYPE", msg)
    #    return
    #  end

    #  q = ttab[type_handle]
    #  set_priority_queue(rsrc, q, pri)
    #endfunction


    #// Function: set_priority_name
    #//
    #// Change the priority of the ~rsrc~ based on the value of ~pri~, the
    #// priority enum argument.  This function changes the priority only in
    #// the name map, leaving the type map untouched.

    def set_priority_name(self, rsrc, pri):
        """
        #function void set_priority_name(uvm_resource_base rsrc,
                                       uvm_resource_types::priority_e pri)
        Args:
            rsrc:
            pri:
        """
        name = ""
        msg = ""
        q = []

        if(rsrc is None):
            uvm_report_warning("NoneRASRC", "attempting to change the serach priority of a None resource")
            return

        name = rsrc.get_name()
        if not self.rtab.exists(name):
            msg = sv.sformatf("Res named %s not found in name map; cannot change its search priority", name)
            uvm_report_error("RNFNAME", msg)
            return


        q = self.rtab[name]
        self.set_priority_queue(rsrc, q, pri)

    #endfunction


    #// Function: set_priority
    #//
    #// Change the search priority of the ~rsrc~ based on the value of ~pri~,
    #// the priority enum argument.  This function changes the priority in
    #// both the name and type maps.

    #function void set_priority (uvm_resource_base rsrc,
    #                            uvm_resource_types::priority_e pri)
    #  set_priority_type(rsrc, pri)
    #  set_priority_name(rsrc, pri)
    #endfunction

    #//--------------------------------------------------------------------
    #// Group: Debug
    #//--------------------------------------------------------------------

    #// Function: find_unused_resources
    #//
    #// Locate all the resources that have at least one write and no reads

    #function uvm_resource_types::rsrc_q_t find_unused_resources()

    #  uvm_resource_types::rsrc_q_t rq
    #  uvm_resource_types::rsrc_q_t q = new
    #  int unsigned i
    #  uvm_resource_base r
    #  uvm_resource_types::access_t a
    #  int reads
    #  int writes

    #  foreach (rtab[name]):
    #    rq = rtab[name]
    #    for(int i=0; i<rq.size(); ++i):
    #      r = rq.get(i)
    #      reads = 0
    #      writes = 0
    #      foreach(r.access[str]):
    #        a = r.access[str]
    #        reads += a.read_count
    #        writes += a.write_count
    #      end
    #      if(writes > 0 && reads == 0)
    #        q.push_back(r)
    #    end
    #  end
    #  return q
    #endfunction


    #// Function: print_resources

    def print_resources(self, rq, audit=False):
        """

        Print the resources that are in a single queue, `rq`.  This is a utility
        function that can be used to print any collection of resources
        stored in a queue.  The `audit` flag determines whether or not the
        audit trail is printed for each resource along with the name,
        value, and scope regular expression.
        #function void print_resources(uvm_resource_types::rsrc_q_t rq, bit audit = 0)
        Args:
            rq:
            audit:
        """
        i = 0
        r = None  # uvm_resource_base r
        UVMResourcePool.printer = UVMLinePrinter()
        printer = UVMResourcePool.printer
        printer.knobs.separator = ""
        printer.knobs.full_name = False
        printer.knobs.identifier = False
        printer.knobs.type_name = False
        printer.knobs.reference = False

        if rq is None or len(rq) == 0:
            uvm_info("UVM/RESOURCE/PRINT","<none>",UVM_NONE)
            return

        for i in range(len(rq)):
            r = rq[i]
            r.print_obj(printer)
            if audit is True:
                r.print_accessors()


    #// Function: dump
    #//
    #// dump the entire resource pool.  The resource pool is traversed and
    #// each resource is printed.  The utility function print_resources()
    #// is used to initiate the printing. If the ~audit~ bit is set then
    #// the audit trail is dumped for each resource.

    def dump(self, audit=False):
        uvm_info("UVM/RESOURCE/DUMP","\n=== resource pool ===",UVM_NONE)
        for name in self.rtab:
            rq = self.rtab[name]
            self.print_resources(rq, audit)
        uvm_info("UVM/RESOURCE/DUMP","=== end of resource pool ===",UVM_NONE)




import unittest
from uvm.base.uvm_object_globals import (
    UVM_DEC, UVM_BIN, UVM_HEX, UVM_OCT)
from uvm.base.uvm_misc import (
    uvm_leaf_scope,
    uvm_bitstream_to_string)


class TestUVMMisc(unittest.TestCase):

    def test_uvm_leaf_scope(self):
        scope = uvm_leaf_scope("[ab]", "[")
        self.assertEqual(scope, "ab")
        scope = uvm_leaf_scope("[xx][ee][ab][cd]", "[")
        self.assertEqual(scope, "cd")
        #scope = uvm_leaf_scope("ab")
        scope = uvm_leaf_scope("a.b.leaf")
        self.assertEqual(scope, "leaf")
        scope = uvm_leaf_scope("leaf")
        self.assertEqual(scope, "leaf")
        #scope = uvm_leaf_scope("a#b#leaf", "#")
        #self.assertEqual(scope, "leaf")

    def test_uvm_bitstream_to_string(self):
        radix = [UVM_DEC, UVM_BIN, UVM_HEX, UVM_OCT]
        for rr in radix:
            num_str = uvm_bitstream_to_string(255, 8, rr)
            if rr == UVM_DEC:
                self.assertEqual(num_str, "{}".format(255))
            elif rr == UVM_HEX:
                self.assertEqual(num_str, "{:X}".format(255))
            elif rr == UVM_BIN:
                self.assertEqual(num_str, "{:b}".format(255))

        num_str = uvm_bitstream_to_string(-245, 8, UVM_DEC)
        self.assertEqual(num_str, "-245")


if __name__ == '__main__':
    unittest.main()




import cocotb
from cocotb.triggers import Timer, Join

from uvm.tlm1.uvm_ports import *
from uvm.tlm1.uvm_imps import *
from uvm.base.uvm_component import UVMComponent
from uvm.base.uvm_event import UVMEvent
from uvm.macros.uvm_object_defines import *
from uvm.seq.uvm_sequence_item import UVMSequenceItem


class UVMTestProducer(UVMComponent):

    def __init__(self, name, parent):
        UVMComponent.__init__(self, name, parent)
        self.put_b_port = UVMBlockingPutPort('put_b_port', self)
        self.get_b_port = UVMBlockingGetPort('get_b_port', self)
        self.put_port = UVMPutPort('put_port', self)
        self.get_port = UVMGetPort('get_port', self)


    async def run_phase(self, phase):
        for i in range(0, 10):
            seq_item = UVMSequenceItem('item')
            exp_id = 2 * i + 1
            seq_item.set_transaction_id(exp_id)
            await self.put_b_port.put(seq_item)
            await Timer(1)
            arr = []
            await self.get_b_port.get(arr)
            got_id = arr[0].get_transaction_id()
            if got_id != exp_id:
                raise Exception("Item ID error. Got: {}, Exp: {}".format(got_id,
                    exp_id))
uvm_component_utils(UVMTestProducer)


class UVMTestConsumer(UVMComponent):

    def __init__(self, name, parent):
        UVMComponent.__init__(self, name, parent)
        self.put_b_imp = UVMBlockingPutImp('put_b_imp', self)
        self.count = 0
        self.wait_n = 10
        self.my_event = UVMEvent('wait data')
        self.get_b_imp = UVMBlockingGetImp('get_b_imp', self)
        self.curr_item = None
        self.get_imp = UVMGetImp('get_imp', self)
        self.put_imp = UVMGetImp('put_imp', self)


    async def put(self, item):
        self.count += 1
        if self.count % 2 == 0:
            await Timer(1)
        else:
            await Timer(2)
        self.curr_item = item
        if self.count == self.wait_n:
            self.my_event.trigger(self.count)


    async def get(self, item):
        await Timer(12)
        item.append(self.curr_item)


    async def wait_all_data(self, n=10):
        self.wait_n = n
        await self.my_event.wait_trigger()


uvm_component_utils(UVMTestConsumer)


class UVMTestExporter(UVMComponent):

    def __init__(self, name, parent):
        UVMComponent.__init__(self, name, parent)
        self.put_export = UVMPutExport('put_export', self)
        #self.fifo =
        self.consumer = UVMTestConsumer('child_cons', self)

    def connect_phase(self, phase):
        self.put_export.connect(self.consumer.put_imp)


    async def wait_all_data(self, n = 10):
        await consumer.wait_all_data(n)


uvm_component_utils(UVMTestExporter)



#//
#//------------------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2013 NVIDIA Corporation
#//   Copyright 2019-2020 Tuomas Poikela
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//------------------------------------------------------------------------------

from .uvm_object import UVMObject
from .uvm_pool import UVMPool
from .uvm_object_globals import (UVM_CALL_HOOK, UVM_COUNT, UVM_DISPLAY, UVM_ERROR, UVM_EXIT,
                                 UVM_FATAL, UVM_INFO, UVM_LOG, UVM_MEDIUM, UVM_NO_ACTION,
                                 UVM_RM_RECORD, UVM_STOP, UVM_WARNING)
from .uvm_globals import uvm_report_enabled
from ..macros.uvm_object_defines import uvm_object_utils

from typing import TextIO, Dict, Union



class UVMReportHandler(UVMObject):
    """
    The UVMReportHandler is the class to which most methods in
    `UVMReportObject` delegate. It stores the maximum verbosity, actions,
    and files that affect the way reports are handled.

    The report handler is not intended for direct use. See `UVMReportObject`
    for information on the UVM reporting mechanism.

    The relationship between `UVMReportObject` (a base class for UVMComponent)
    and UVMReportHandler is typically one to one, but it can be many to one
    if several uvm_report_objects are configured to use the same
    `UVMReportHandler` object. See `UVMReportObject.set_report_handler`.

    The relationship between UVMReportHandler and `UVMReportServer` is many
    to one.
    """

    def __init__(self, name=""):
        UVMObject.__init__(self, name)
        self.m_max_verbosity_level = 0

        # id verbosity settings : default and severity
        self.id_verbosities = UVMPool()
        self.severity_id_verbosities: Dict[str, UVMPool] = {}  # string -> UVMPool()

        # actions
        self.id_actions = UVMPool()
        self.severity_actions = UVMPool()
        self.severity_id_actions: Dict[str, UVMPool] = {}  # string -> UVMPool

        # severity overrides
        self.sev_overrides = UVMPool()
        self.sev_id_overrides: Dict[str, UVMPool] = {}  # string -> UVMPool

        # file handles : default, severity, action, (severity,id)
        self.default_file_handle: Union[int, TextIO] = 1
        self.id_file_handles = UVMPool()
        self.severity_file_handles = {}  # severity -> UVM_FILE
        self.severity_id_file_handles = {}  # severity -> UVMPool
        self.initialize()

    def do_print(self, printer):
        """
        The uvm_report_handler implements the <uvm_object::do_print()> such that
        `print` method provides UVM printer formatted output
        of the current configuration.  A snippet of example output is shown
        here::

          uvm_test_top                uvm_report_handler  -     @555
            max_verbosity_level       uvm_verbosity       32    UVM_FULL
            id_verbosities            uvm_pool            3     -
              [ID1]                   uvm_verbosity       32    UVM_LOW
            severity_id_verbosities   array               4     -
              [UVM_INFO:ID4]          int                 32    501
            id_actions                uvm_pool            2     -
              [ACT_ID]                uvm_action          32    DISPLAY LOG COUNT
            severity_actions          array               4     -
              [UVM_INFO]              uvm_action          32    DISPLAY
              [UVM_WARNING]           uvm_action          32    DISPLAY RM_RECORD COUNT
              [UVM_ERROR]             uvm_action          32    DISPLAY COUNT
              [UVM_FATAL]             uvm_action          32    DISPLAY EXIT
            default_file_handle       int                 32    'h1

        Args:
            printer:
        """
        print("UVMReportHandler do_print() TODO")
        #  virtual def void do_print (self,uvm_printer printer):
        #
        #    uvm_verbosity l_verbosity
        #    uvm_severity l_severity
        #    string idx
        #    int l_int
        #
        #    // max verb
        #    if (sv.cast(l_verbosity, m_max_verbosity_level))
        #      printer.print_generic("max_verbosity_level", "uvm_verbosity", 32,
        #        l_verbosity.name())
        #    else
        #      printer.print_int("max_verbosity_level", m_max_verbosity_level, 32, UVM_DEC,
        #        ".", "int")
        #
        #    // id verbs
        #    if(id_verbosities.first(idx)):
        #      printer.print_array_header("id_verbosities",id_verbosities.num(),
        #        "uvm_pool")
        #      do begin
        #        l_int = id_verbosities.get(idx)
        #        if (sv.cast(l_verbosity, l_int))
        #          printer.print_generic(sv.sformatf("[%s]", idx), "uvm_verbosity", 32,
        #            l_verbosity.name())
        #        else begin
        #          string l_str
        #          l_str.itoa(l_int)
        #          printer.print_generic(sv.sformatf("[%s]", idx), "int", 32,
        #            l_str)
        #        end
        #      end while(id_verbosities.next(idx))
        #      printer.print_array_footer()
        #    end
        #
        #    // sev and id verbs
        #    if(severity_id_verbosities.size() != 0):
        #      int _total_cnt
        #      foreach (severity_id_verbosities[l_severity])
        #        _total_cnt += severity_id_verbosities[l_severity].num()
        #      printer.print_array_header("severity_id_verbosities", _total_cnt,
        #        "array")
        #      if(severity_id_verbosities.first(l_severity)):
        #        do begin
        #          uvm_id_verbosities_array id_v_ary = severity_id_verbosities[l_severity]
        #          if(id_v_ary.first(idx))
        #          do begin
        #            l_int = id_v_ary.get(idx)
        #            if (sv.cast(l_verbosity, l_int))
        #              printer.print_generic(sv.sformatf("[%s:%s]", l_severity.name(), idx),
        #				    "uvm_verbosity", 32, l_verbosity.name())
        #            else begin
        #              string l_str
        #              l_str.itoa(l_int)
        #              printer.print_generic(sv.sformatf("[%s:%s]", l_severity.name(), idx),
        #				    "int", 32, l_str)
        #            end
        #          end while(id_v_ary.next(idx))
        #        end while(severity_id_verbosities.next(l_severity))
        #      end
        #      printer.print_array_footer()
        #    end
        #
        #    // id actions
        #    if(id_actions.first(idx)):
        #      printer.print_array_header("id_actions",id_actions.num(),
        #        "uvm_pool")
        #      do begin
        #        l_int = id_actions.get(idx)
        #        printer.print_generic(sv.sformatf("[%s]", idx), "uvm_action", 32,
        #          format_action(l_int))
        #      end while(id_actions.next(idx))
        #      printer.print_array_footer()
        #    end
        #
        #    // severity actions
        #    if(severity_actions.first(l_severity)):
        #      printer.print_array_header("severity_actions",4,"array")
        #      do begin
        #        printer.print_generic(sv.sformatf("[%s]", l_severity.name()), "uvm_action", 32,
        #          format_action(severity_actions[l_severity]))
        #      end while(severity_actions.next(l_severity))
        #      printer.print_array_footer()
        #    end
        #
        #    // sev and id actions
        #    if(severity_id_actions.size() != 0):
        #      int _total_cnt
        #      foreach (severity_id_actions[l_severity])
        #        _total_cnt += severity_id_actions[l_severity].num()
        #      printer.print_array_header("severity_id_actions", _total_cnt,
        #        "array")
        #      if(severity_id_actions.first(l_severity)):
        #        do begin
        #          uvm_id_actions_array id_a_ary = severity_id_actions[l_severity]
        #          if(id_a_ary.first(idx))
        #          do begin
        #            printer.print_generic(sv.sformatf("[%s:%s]", l_severity.name(), idx),
        #				  "uvm_action", 32, format_action(id_a_ary.get(idx)))
        #          end while(id_a_ary.next(idx))
        #        end while(severity_id_actions.next(l_severity))
        #      end
        #      printer.print_array_footer()
        #    end
        #
        #    // sev overrides
        #    if(sev_overrides.first(l_severity)):
        #      printer.print_array_header("sev_overrides",sev_overrides.num(),
        #        "uvm_pool")
        #      do begin
        #        uvm_severity l_severity_new = sev_overrides.get(l_severity)
        #        printer.print_generic(sv.sformatf("[%s]", l_severity.name()),
        #          "uvm_severity", 32, l_severity_new.name())
        #      end while(sev_overrides.next(l_severity))
        #      printer.print_array_footer()
        #    end
        #
        #    // sev and id overrides
        #    if(sev_id_overrides.size() != 0):
        #      int _total_cnt
        #      foreach (sev_id_overrides[idx])
        #        _total_cnt += sev_id_overrides[idx].num()
        #      printer.print_array_header("sev_id_overrides", _total_cnt,
        #        "array")
        #      if(sev_id_overrides.first(idx)):
        #        do begin
        #          uvm_sev_override_array sev_o_ary = sev_id_overrides[idx]
        #          if(sev_o_ary.first(l_severity))
        #          do begin
        #            uvm_severity new_sev = sev_o_ary.get(l_severity)
        #            printer.print_generic(sv.sformatf("[%s:%s]", l_severity.name(), idx),
        #              "uvm_severity", 32, new_sev.name())
        #          end while(sev_o_ary.next(l_severity))
        #        end while(sev_id_overrides.next(idx))
        #      end
        #      printer.print_array_footer()
        #    end
        #
        #    // default file handle
        #    printer.print_int("default_file_handle", default_file_handle, 32, UVM_HEX,
        #      ".", "int")
        #
        #    // id files
        #    if(id_file_handles.first(idx)):
        #      printer.print_array_header("id_file_handles",id_file_handles.num(),
        #        "uvm_pool")
        #      do begin
        #        printer.print_int(sv.sformatf("[%s]", idx), id_file_handles.get(idx), 32,
        #          UVM_HEX, ".", "UVM_FILE")
        #      end while(id_file_handles.next(idx))
        #      printer.print_array_footer()
        #    end
        #
        #    // severity files
        #    if(severity_file_handles.first(l_severity)):
        #      printer.print_array_header("severity_file_handles",4,"array")
        #      do begin
        #        printer.print_int(sv.sformatf("[%s]", l_severity.name()),
        #          severity_file_handles[l_severity], 32, UVM_HEX, ".", "UVM_FILE")
        #      end while(severity_file_handles.next(l_severity))
        #      printer.print_array_footer()
        #    end
        #
        #    // sev and id files
        #    if(severity_id_file_handles.size() != 0):
        #      int _total_cnt
        #      foreach (severity_id_file_handles[l_severity])
        #        _total_cnt += severity_id_file_handles[l_severity].num()
        #      printer.print_array_header("severity_id_file_handles", _total_cnt,
        #        "array")
        #      if(severity_id_file_handles.first(l_severity)):
        #        do begin
        #          uvm_id_file_array id_f_ary = severity_id_file_handles[l_severity]
        #          if(id_f_ary.first(idx))
        #          do begin
        #            printer.print_int(sv.sformatf("[%s:%s]", l_severity.name(), idx),
        #              id_f_ary.get(idx), 32, UVM_HEX, ".", "UVM_FILE")
        #          end while(id_f_ary.next(idx))
        #        end while(severity_id_file_handles.next(l_severity))
        #      end
        #      printer.print_array_footer()
        #    end
        #
        #  endfunction


    #  //----------------------------------------------------------------------------
    #  // Group: Message Processing
    #  //----------------------------------------------------------------------------


    def process_report_message(self, report_message):
        """
        This is the common handler method used by the four core reporting methods
        (e.g. `uvm_report_error`) in `uvm_report_object`.

        Args:
            report_message (UVMReportMessage): Message to process
        """
        from .uvm_report_server import UVMReportServer
        srvr = UVMReportServer.get_server()
        id = report_message.get_id()
        severity = report_message.get_severity()

        # Check for severity overrides and apply them before calling the server.
        # An id specific override has precedence over a generic severity override.
        if id in self.sev_id_overrides:
            if self.sev_id_overrides[id].exists(severity):
                severity = self.sev_id_overrides[id].get(severity)
                report_message.set_severity(severity)
        else:
            if self.sev_overrides.exists(severity):
                severity = self.sev_overrides.get(severity)
                report_message.set_severity(severity)
        report_message.set_file(self.get_file_handle(severity, id))
        report_message.set_report_handler(self)
        report_message.set_action(self.get_action(severity, id))
        srvr.process_report_message(report_message)

    #  //----------------------------------------------------------------------------
    #  // Group: Convenience Methods
    #  //----------------------------------------------------------------------------

    @classmethod
    def format_action(cls, action):
        """
          Function: format_action

          Returns a string representation of the `action`, e.g., "DISPLAY".
        Args:
            cls:
            action:
        Returns:
        """
        s = ""
        if action == UVM_NO_ACTION:
            s = "UVM_NO_ACTION"
        else:
            if action & UVM_DISPLAY:
                s = s + "DISPLAY "
            if action & UVM_LOG:
                s = s + "LOG "
            if action & UVM_RM_RECORD:
                s = s + "RM_RECORD "
            if action & UVM_COUNT:
                s = s + "COUNT "
            if action & UVM_CALL_HOOK:
                s = s + "CALL_HOOK "
            if action & UVM_EXIT:
                s = s + "EXIT "
            if action & UVM_STOP:
                s = s + "STOP "
        return s

    def initialize(self):
        """
          Function- initialize

          Internal method for initializing report handler.
        """
        self.set_default_file(0)
        self.m_max_verbosity_level = UVM_MEDIUM
        self.set_severity_action(UVM_INFO, UVM_DISPLAY)
        self.set_severity_action(UVM_WARNING, UVM_DISPLAY)
        self.set_severity_action(UVM_ERROR, UVM_DISPLAY | UVM_COUNT)
        self.set_severity_action(UVM_FATAL, UVM_DISPLAY | UVM_EXIT)

        self.set_severity_file(UVM_INFO, self.default_file_handle)
        self.set_severity_file(UVM_WARNING, self.default_file_handle)
        self.set_severity_file(UVM_ERROR, self.default_file_handle)
        self.set_severity_file(UVM_FATAL, self.default_file_handle)

    def get_severity_id_file(self, severity, id):
        """
          Function- get_severity_id_file

          Return the file id based on the severity and the id
        Args:
            severity:
            id:
        Returns:
        """
        """ Return the file id based on the severity and the id """
        if severity in self.severity_id_file_handles:
            array = self.severity_id_file_handles[severity]
            if array.exists(id):
                return array.get(id)

        if self.id_file_handles.exists(id):
            return self.id_file_handles.get(id)

        if severity in self.severity_file_handles:
            return self.severity_file_handles[severity]
        return self.default_file_handle

    def set_verbosity_level(self, verbosity_level: int):
        """
        Internal method called by uvm_report_object.
        Args:
            verbosity_level:
        """
        self.m_max_verbosity_level = verbosity_level

    def get_verbosity_level(self, severity=UVM_INFO, id="") -> int:
        """
        Returns the verbosity associated with the given `severity` and `id`.

        First, if there is a verbosity associated with the ~(severity,id)~ pair,
        return that.  Else, if there is a verbosity associated with the `id`, return
        that.  Else, return the max verbosity setting.
        Args:
            severity:
            id:
        Returns:
            int: Verbosity level of this handler.
        """
        if severity in self.severity_id_verbosities:
            array = self.severity_id_verbosities[severity]
            if array.exists(id):
                return array.get(id)

        if self.id_verbosities.exists(id):
            return self.id_verbosities.get(id)

        return self.m_max_verbosity_level

    def get_action(self, severity, id):
        """
        Function- get_action

        Returns the action associated with the given `severity` and `id`.

        First, if there is an action associated with the ~(severity,id)~ pair,
        return that.  Else, if there is an action associated with the `id`, return
        that.  Else, if there is an action associated with the `severity`, return
        that. Else, return the default action associated with the `severity`.
        Args:
            severity:
            id:
        Returns:
        """
        if severity in self.severity_id_actions:
            array = self.severity_id_actions[severity]
            if array.exists(id):
                return array.get(id)

        if self.id_actions.exists(id):
            return self.id_actions.get(id)
        return self.severity_actions.get(severity)

    def get_file_handle(self, severity, id):
        """
        Function- get_file_handle

        Returns the file descriptor associated with the given `severity` and `id`.

        First, if there is a file handle associated with the ~(severity,id)~ pair,
        return that. Else, if there is a file handle associated with the `id`, return
        that. Else, if there is an file handle associated with the `severity`, return
        that. Else, return the default file handle.
        Args:
            severity:
            id:
        Returns:
        """
        _file = self.get_severity_id_file(severity, id)
        if _file != 0:
            return _file

        if self.id_file_handles.exists(id):
            _file = self.id_file_handles.get(id)
            if _file != 0:
                return _file

        if severity in self.severity_file_handles:
            _file = self.severity_file_handles[severity]
            if _file != 0:
                return _file
        return self.default_file_handle


    def set_severity_action(self, severity, action):
        """
          Function- set_severity_action
        Args:
            severity:
            action:
        """
        self.severity_actions.add(severity, action)

    def set_id_action(self, id, action):
        """
          Function- set_id_action
        Args:
            id:
            action:
        """
        self.id_actions.add(id, action)

    def set_severity_id_action(self, severity, id, action):
        """
          Function- set_severity_id_action
        Args:
            severity:
            id:
            action:
        """
        if severity not in self.severity_id_actions:
            self.severity_id_actions[severity] = UVMPool()
        self.severity_id_actions[severity].add(id, action)

    def set_id_verbosity(self, id, verbosity):
        """
          Function- set_id_verbosity
        Args:
            id:
            verbosity:
        """
        self.id_verbosities.add(id, verbosity)

    def set_severity_id_verbosity(self, severity, id, verbosity):
        """
          Function- set_severity_id_verbosity
        Args:
            severity:
            id:
            verbosity:
        """
        if severity not in self.severity_id_verbosities:
            self.severity_id_verbosities[severity] = UVMPool()
        self.severity_id_verbosities[severity].add(id,verbosity)

    def set_default_file(self, file):
        """
          Function- set_default_file
        Args:
            file:
        """
        self.default_file_handle = file

    def set_severity_file(self, severity, file):
        """
          Function- set_severity_file
        Args:
            severity:
            file:
        """
        self.severity_file_handles[severity] = file

    def set_id_file(self, id, file):
        """
          Function- set_id_file
        Args:
            id:
            file:
        """
        self.id_file_handles.add(id, file)

    def set_severity_id_file(self, severity, id, file):
        """
          Function- set_severity_id_file
        Args:
            severity:
            id:
            file:
        """
        if severity not in self.severity_id_file_handles:
            self.severity_id_file_handles[severity] = UVMPool()
        self.severity_id_file_handles[severity].add(id, file)

    def set_severity_override(self, cur_severity, new_severity):
        self.sev_overrides.add(cur_severity, new_severity)

    def set_severity_id_override(self, cur_severity, id, new_severity):
        # has precedence over set_severity_override
        # silently override previous setting
        if id not in self.sev_id_overrides:
            self.sev_id_overrides[id] = UVMPool()
        self.sev_id_overrides[id].add(cur_severity, new_severity)

    def report(self, severity, name, id, message,
      verbosity_level=UVM_MEDIUM, filename="", line=0,
      client=None):
        """
        Function- report

        This is the common handler method used by the four core reporting methods
        (e.g., uvm_report_error) in `uvm_report_object`.
        Args:
            severity:
            name:
            id:
            message:
            verbosity_level:
            filename:
            line:
            client:
        """

        l_report_enabled = False
        l_report_message = None
        from .uvm_coreservice import UVMCoreService
        cs = UVMCoreService.get()
        if not uvm_report_enabled(verbosity_level, UVM_INFO, id):
            return

        if client is None:
            client = cs.get_root()

        from .uvm_report_message import UVMReportMessage
        l_report_message = UVMReportMessage.new_report_message()
        l_report_message.set_report_message(severity, id, message,
                                            verbosity_level, filename, line, name)
        l_report_message.set_report_object(client)
        l_report_message.set_action(self.get_action(severity, id))
        self.process_report_message(l_report_message)


    def _close_files(self):
        if self.default_file_handle != 0:
            if hasattr(self.default_file_handle, 'close'):
                self.default_file_handle.close()
        for sev in self.severity_file_handles:
            if self.severity_file_handles[sev] != 0:
                if hasattr(self.severity_file_handles[sev], 'close'):
                    self.severity_file_handles[sev].close()

        #for id in self.severity_id_file_handles[severity].add(id, file)

uvm_object_utils(UVMReportHandler)



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2009 Synopsys, Inc.
#//    Copyright 2010 Mentor Graphics Corporation
#//    Copyright 2019 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
"""
Title: Classes for Adapting Between Register and Bus Operations

This section defines classes used to convert transaction streams between
generic register address/data reads and writes and physical bus accesses.

Group: Example

The following example illustrates how to implement a RegModel-BUS adapter class
for the APB bus protocol::

    class rreg2apb_adapter(UVMRegAdapter):

        def __init__(self, name="reg2apb_adapter"):
            super().__init__(name)

        def reg2bus(self, rw) -> UVMSequenceItem:
            apb_item apb = apb_item.type_id.create("apb_item")
            apb.op   = (rw.kind == UVM_READ) ? apb.READ : apb.WRITE
            apb.addr = rw.addr
            apb.data = rw.data
            return apb

        def bus2reg(self, bus_item, rw):
            arr_apb = []
            if not sv.cast(arr_apb, bus_item, apb_item)) begin
              uvm_fatal("CONVERT_APB2REG","Bus item is not of type apb_item")
            end
            apb = arr_apb[0]

            rw.kind  = apb.op==apb.READ ? UVM_READ : UVM_WRITE
            rw.addr = apb.addr
            rw.data = apb.data
            rw.status = UVM_IS_OK

    uvm_object_utils(reg2apb_adapter)
"""

from ..base.uvm_object import UVMObject
from ..macros.uvm_message_defines import *



#//------------------------------------------------------------------------------
#//
#// Class: uvm_reg_adapter
#//
#// This class defines an interface for converting between <uvm_reg_bus_op>
#// and a specific bus transaction.
#//------------------------------------------------------------------------------


class UVMRegAdapter(UVMObject):

    def __init__(self, name=""):
        """
          Function: new

          Create a new instance of this type, giving it the optional `name`.
        Args:
            name:
        """
        UVMObject.__init__(self, name)
        self.supports_byte_enable = False
        self.provides_responses = False
        self.parent_sequence = None
        self.m_item = None  # uvm_reg_item m_item

    #  // Variable: supports_byte_enable
    #  //
    #  // Set this bit in extensions of this class if the bus protocol supports
    #  // byte enables.
    #
    #  bit supports_byte_enable

    #  // Variable: provides_responses
    #  //
    #  // Set this bit in extensions of this class if the bus driver provides
    #  // separate response items.
    #
    #  bit provides_responses

    #  // Variable: parent_sequence
    #  //
    #  // Set this member in extensions of this class if the bus driver requires
    #  // bus items be executed via a particular sequence base type. The sequence
    #  // assigned to this member must implement do_clone().
    #
    #  uvm_sequence_base parent_sequence

    def reg2bus(self, rw):
        """
          Function: reg2bus

          Extensions of this class `must` implement this method to convert the specified
          `uvm_reg_bus_op` to a corresponding `uvm_sequence_item` subtype that defines the bus
          transaction.

          The method must allocate a new bus-specific `uvm_sequence_item`,
          assign its members from
          the corresponding members from the given generic `rw` bus operation, then
          return it.

         pure virtual function uvm_sequence_item reg2bus(const ref uvm_reg_bus_op rw)
        Args:
            rw:
        Raises:
        """
        raise NotImplementedError("Pure virtual function")


    def bus2reg(self, bus_item, rw):
        """
          Function: bus2reg

          Extensions of this class `must` implement this method to copy members
          of the given bus-specific `bus_item` to corresponding members of the provided
          `bus_rw` instance. Unlike `reg2bus`, the resulting transaction
          is not allocated from scratch. This is to accommodate applications
          where the bus response must be returned in the original request.

         pure virtual function void bus2reg(uvm_sequence_item bus_item,
                                            ref uvm_reg_bus_op rw)


        Args:
            bus_item:
            rw:
        Raises:
        """
        raise NotImplementedError("Pure virtual function")

    #  // function: get_item
    #  //
    #  // Returns the bus-independent read/write information that corresponds to
    #  // the generic bus transaction currently translated to a bus-specific
    #  // transaction.
    #  // This function returns a value reference only when called in the
    #  // <uvm_reg_adapter::reg2bus()> method.
    #  // It returns ~null~ at all other times.
    #  // The content of the return <uvm_reg_item> instance must not be modified
    #  // and used strictly to obtain additional information about the operation.
    def get_item(self):
        return self.m_item


    def m_set_item(self, item):
        """
         virtual function void m_set_item(uvm_reg_item item)
        Args:
            item:
        """
        if item is not None and isinstance(item.offset, int) is False:
            uvm_fatal("OFFSET/NOT_INT", "Item offset must be int. Got: {}"
                .format(item.offset))

        self.m_item = item




#//------------------------------------------------------------------------------
#//
#// Class: uvm_reg_tlm_adapter
#//
#// For converting between <uvm_reg_bus_op> and <uvm_tlm_gp> items.
#//
#//------------------------------------------------------------------------------
#

#class uvm_reg_tlm_adapter extends uvm_reg_adapter
    #
    #  `uvm_object_utils(uvm_reg_tlm_adapter)
    #
    #  function new(string name = "uvm_reg_tlm_adapter")
    #    super.new(name)
    #  endfunction
    #
    #  // Function: reg2bus
    #  //
    #  // Converts a <uvm_reg_bus_op> struct to a <uvm_tlm_gp> item.
    #
    #  virtual function uvm_sequence_item reg2bus(const ref uvm_reg_bus_op rw)
    #
    #     uvm_tlm_gp gp = uvm_tlm_gp::type_id::create("tlm_gp",, this.get_full_name())
    #     int nbytes = (rw.n_bits-1)/8+1
    #     uvm_reg_addr_t addr=rw.addr
    #
    #     if (rw.kind == UVM_WRITE)
    #        gp.set_command(UVM_TLM_WRITE_COMMAND)
    #     else
    #        gp.set_command(UVM_TLM_READ_COMMAND)
    #
    #     gp.set_address(addr)
    #
    #     gp.m_byte_enable = new [nbytes]
    #     gp.m_byte_enable_length = nbytes
    #
    #     gp.set_streaming_width (nbytes)
    #
    #     gp.m_data = new [gp.get_streaming_width()]
    #     gp.m_length = nbytes
    #
    #     for (int i = 0; i < nbytes; i++) begin
    #        gp.m_data[i] = rw.data[i*8+:8]
    #        gp.m_byte_enable[i] = (i > nbytes) ? 8'h00 : (rw.byte_en[i] ? 8'hFF : 8'h00)
    #     end
    #
    #     return gp
    #
    #  endfunction
    #
    #
    #  // Function: bus2reg
    #  //
    #  // Converts a <uvm_tlm_gp> item to a <uvm_reg_bus_op>.
    #  // into the provided ~rw~ transaction.
    #  //
    #  virtual function void bus2reg(uvm_sequence_item bus_item,
    #                                ref uvm_reg_bus_op rw)
    #
    #    uvm_tlm_gp gp
    #    int nbytes
    #
    #    if (bus_item == null)
    #     `uvm_fatal("REG/NULL_ITEM","bus2reg: bus_item argument is null")
    #
    #    if (!$cast(gp,bus_item)) begin
    #      `uvm_error("WRONG_TYPE","Provided bus_item is not of type uvm_tlm_gp")
    #      return
    #    end
    #
    #    if (gp.get_command() == UVM_TLM_WRITE_COMMAND)
    #      rw.kind = UVM_WRITE
    #    else
    #      rw.kind = UVM_READ
    #
    #    rw.addr = gp.get_address()
    #
    #    rw.byte_en = 0
    #    foreach (gp.m_byte_enable[i])
    #      rw.byte_en[i] = gp.m_byte_enable[i]
    #
    #    rw.data = 0
    #    foreach (gp.m_data[i])
    #      rw.data[i*8+:8] = gp.m_data[i]
    #
    #    rw.status = (gp.is_response_ok()) ? UVM_IS_OK : UVM_NOT_OK
    #
    #
    #  endfunction
    #
    #endclass
    #



#----------------------------------------------------------------------
#   Copyright 2007-2011 Mentor Graphics Corporation
#   Copyright 2007-2010 Cadence Design Systems, Inc.
#   Copyright 2010 Synopsys, Inc.
#   Copyright 2019 Tuomas Poikela (tpoikela)
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#----------------------------------------------------------------------


#------------------------------------------------------------------------------
#
# Class: uvm_topdown_phase
#
#------------------------------------------------------------------------------
# Virtual base class for function phases that operate top-down.
# The pure virtual function execute() is called for each component.
#
# A top-down function phase completes when the <execute()> method
# has been called and returned on all applicable components
# in the hierarchy.

from .uvm_phase import UVMPhase
from .uvm_globals import uvm_report_fatal, uvm_report_info
from .uvm_debug import uvm_debug
from .uvm_object_globals import (UVM_DEBUG, UVM_PHASE_ENDED, UVM_PHASE_EXECUTING, UVM_PHASE_IMP,
                                 UVM_PHASE_READY_TO_END, UVM_PHASE_STARTED)


class UVMTopdownPhase(UVMPhase):

    def __init__(self, name):
        """
        Function: new

        Create a new instance of a top-down phase

        Args:
            name:
        """
        UVMPhase.__init__(self, name, UVM_PHASE_IMP)

    def traverse(self, comp, phase, state):
        """
        Function: traverse

        Traverses the component tree in top-down order, calling `execute` for
        each component.

        Args:
            comp:
            phase:
            state:
        """
        uvm_debug(self, 'traverse', self.get_name() +
            ' traversing topdown phase now with comp' + comp.get_name())
        name = ""
        phase_domain = phase.get_domain()
        comp_domain = comp.get_domain()

        if UVMPhase.m_phase_trace:
            dom_name = "NO DOMAIN"
            if comp_domain is not None:
                dom_name = comp_domain.get_name()
            uvm_report_info("PH_TRACE", ("topdown-phase phase={} state={} comp={}"
                + "comp.domain={} phase.domain={}").format(
                str(phase), str(state), comp.get_full_name(),
                    dom_name, phase_domain.get_name()), UVM_DEBUG)

        from .uvm_domain import UVMDomain
        if phase_domain == UVMDomain.get_common_domain() or phase_domain == comp_domain:
            if state == UVM_PHASE_STARTED:
                comp.m_current_phase = phase
                comp.m_apply_verbosity_settings(phase)
                comp.phase_started(phase)
            elif state == UVM_PHASE_EXECUTING:
                if not(phase.get_name() == "build" and comp.m_build_done):
                    ph = self
                    comp.m_phasing_active += 1
                    if self in comp.m_phase_imps:
                        ph = comp.m_phase_imps[self]
                    ph.execute(comp, phase)
                    comp.m_phasing_active -= 1
            elif state == UVM_PHASE_READY_TO_END:
                comp.phase_ready_to_end(phase)
            elif state == UVM_PHASE_ENDED:
                comp.phase_ended(phase)
                comp.m_current_phase = None
            else:
                uvm_report_fatal("PH_BADEXEC","topdown phase traverse internal error")

        if comp.has_first_child():
            child = comp.get_first_child()
            while child is not None:
                self.traverse(child, phase, state)
                child = comp.get_next_child()

    def execute(self, comp, phase):
        """
        Function: execute

        Executes the top-down phase `phase` for the component `comp`.

        Args:
            comp:
            phase:
        """
        # reseed this process for random stability
        #process proc = process::self();
        #proc.srandom(uvm_create_random_seed(phase.get_type_name(), comp.get_full_name()));

        comp.m_current_phase = phase
        self.exec_func(comp,phase)



#//----------------------------------------------------------------------
#// Copyright 2010-2011 Mentor Graphics Corporation
#// Copyright 2014 Semifore
#// Copyright 2010 Synopsys, Inc.
#// Copyright 2010-2018 Cadence Design Systems, Inc.
#// Copyright 2014-2015 NVIDIA Corporation
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

import cocotb
from uvm.base.uvm_port_base import UVMPortBase
from ..macros.uvm_message_defines import (uvm_error)

#//----------------------------------------------------------------------
#// Title -- NODOCS -- TLM2 imps (interface implementations)
#//
#// This section defines the implementation classes for connecting TLM2
#// interfaces.
#//
#// TLM imps bind a TLM interface with the object that contains the
#// interface implementation.
#// In addition to the transaction type and the phase type, the imps
#// are parameterized with the type of the object that will provide the
#// implementation. Most often this will be the type of the component
#// where the imp resides. The constructor of the imp takes as an argument
#// an object of type IMP and installs it as the implementation object.
#// Most often the imp constructor argument is "this".
#//----------------------------------------------------------------------

#//--------------------------
#// Group -- NODOCS -- IMP binding macros
#//--------------------------

from uvm.tlm2.uvm_tlm2_ifs import uvm_tlm_sync_e
from uvm.base.uvm_port_base import UVMPortBase
from uvm.tlm1.uvm_tlm_imps import UVM_IMP_COMMON
from uvm.tlm2.uvm_tlm2_defines import UVM_TLM_NB_FW_MASK, UVM_TLM_B_MASK,\
    UVM_TLM_NB_BW_MASK
from uvm.macros.uvm_message_defines import uvm_error
from uvm.base.sv import cat
import cocotb

#// The macro wraps the forward path call function nb_transport_fw()
#//
#// The first call to this method for a transaction marks the initial timing point.
#// Every call to this method may mark a timing point in the execution of the
#// transaction. The timing annotation argument allows the timing points
#// to be offset from the simulation times at which the forward path is used.
#// The final timing point of a transaction may be marked by a call
#// to nb_transport_bw() within <`UVM_TLM_NB_TRANSPORT_BW_IMP> or a return from this
#// or subsequent call to nb_transport_fw().
#//
#// See <TLM2 Interfaces, Ports, Exports and Transport Interfaces Subset>
#// for more details on the semantics and rules of the nonblocking
#// transport interface.
def UVM_TLM_NB_TRANSPORT_FW_IMP(imp, T):
    def nb_transport_fw(self, t, p, delay):
        if delay is None:
            uvm_error("UVM/TLM/NULLDELAY",
                cat(self.get_full_name(),
                   ".nb_transport_fw() called with 'null' delay"))
            return uvm_tlm_sync_e.UVM_TLM_COMPLETED

        return getattr(self, imp).nb_transport_fw(t, p, delay)
    setattr(T, "nb_transport_fw", nb_transport_fw)


#// Implementation of the backward path.
#// The macro wraps the function called nb_transport_bw().
#// This function MUST be implemented in the INITIATOR component class.
#//
#// Every call to this method may mark a timing point, including the final
#// timing point, in the execution of the transaction.
#// The timing annotation argument allows the timing point
#// to be offset from the simulation times at which the backward path is used.
#// The final timing point of a transaction may be marked by a call
#// to nb_transport_fw() within <`UVM_TLM_NB_TRANSPORT_FW_IMP> or a return from
#// this or subsequent call to nb_transport_bw().
#//
#// See <TLM2 Interfaces, Ports, Exports and Transport Interfaces Subset>
#// for more details on the semantics and rules of the nonblocking
#// transport interface.
#//
#// Example:
#//
#//| class master extends uvm_component;
#//|    uvm_tlm_nb_initiator_socket
#//|          #(trans, uvm_tlm_phase_e, this_t) initiator_socket;
#//|
#//|    function void build_phase(uvm_phase phase);
#//|       initiator_socket = new("initiator_socket", this, this);
#//|    endfunction
#//|
#//|    function uvm_tlm_sync_e nb_transport_bw(trans t,
#//|                                   ref uvm_tlm_phase_e p,
#//|                                   input uvm_tlm_time delay);
#//|        transaction = t;
#//|        state = p;
#//|        return UVM_TLM_ACCEPTED;
#//|    endfunction
#//|
#//|    ...
#//| endclass
def UVM_TLM_NB_TRANSPORT_BW_IMP(imp, T):
    def nb_transport_bw(self, t, p, delay):
        if delay is None:
            uvm_error("UVM/TLM/NULLDELAY",
                cat(self.get_full_name(),
                   ".nb_transport_bw() called with 'null' delay"))
            return uvm_tlm_sync_e.UVM_TLM_COMPLETED
        return getattr(self, imp).nb_transport_bw(t, p, delay)

    setattr(T, "nb_transport_bw", nb_transport_bw)


#// The macro wraps the function b_transport()
#// Execute a blocking transaction. Once this method returns,
#// the transaction is assumed to have been executed. Whether
#// that execution is successful or not must be indicated by the
#// transaction itself.
#//
#// The callee may modify or update the transaction object, subject
#// to any constraints imposed by the transaction class. The
#// initiator may re-use a transaction object from one call to
#// the next and across calls to b_transport().
#//
#// The call to b_transport shall mark the first timing point of the
#// transaction. The return from b_transport() shall mark the final
#// timing point of the transaction. The timing annotation argument
#// allows the timing points to be offset from the simulation times
#// at which the task call and return are executed.
def UVM_TLM_B_TRANSPORT_IMP(imp, T):

    async def b_transport(self, t, delay):
        if delay == None:
            uvm_error("UVM/TLM/NULLDELAY",
                  cat(self.get_full_name(),
                   ".b_transport() called with 'null' delay"))
            return
        await getattr(self, imp).b_transport(t, delay)
    setattr(T, "b_transport", b_transport)


#//---------------------------
#// Group -- NODOCS -- IMP binding classes
#//---------------------------

#//
#// Used like exports, except an additional class parameter specifies
#// the type of the implementation object.  When the
#// imp is instantiated the implementation object is bound.
class UVMTLMBTransportImp(UVMPortBase):
    pass

UVMTLMBTransportImp = UVM_IMP_COMMON(UVMTLMBTransportImp,  # type: ignore
        UVM_TLM_B_MASK, "uvm_tlm_b_transport_imp")
UVM_TLM_B_TRANSPORT_IMP('m_imp', UVMTLMBTransportImp)


#// Used like exports, except an additional class parameter specifies
#// the type of the implementation object.  When the
#// imp is instantiated the implementation object is bound.
class UVMTLMNbTransportFwImp(UVMPortBase):
    pass

UVMTLMNbTransportFwImp = UVM_IMP_COMMON(UVMTLMNbTransportFwImp,  # type: ignore
        UVM_TLM_NB_FW_MASK, "uvm_tlm_nb_transport_fw_imp")
UVM_TLM_NB_TRANSPORT_FW_IMP('m_imp', UVMTLMNbTransportFwImp)


#// Used like exports, except an additional class parameter specifies
#// the type of the implementation object.  When the
#// imp is instantiated the implementation object is bound.
class UVMTLMNbTransportBwImp(UVMPortBase):
    pass

UVMTLMNbTransportBwImp = UVM_IMP_COMMON(UVMTLMNbTransportBwImp,  # type: ignore
        UVM_TLM_NB_BW_MASK, "uvm_tlm_nb_transport_bw_imp")
UVM_TLM_NB_TRANSPORT_BW_IMP('m_imp', UVMTLMNbTransportBwImp)




#----------------------------------------------------------------------
#   Copyright 2007-2010 Mentor Graphics Corporation
#   Copyright 2007-2011 Cadence Design Systems, Inc.
#   Copyright 2010-2011 Synopsys, Inc.
#   Copyright 2019 Tuomas Poikela (tpoikela)
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#----------------------------------------------------------------------

import cocotb
from cocotb.triggers import Timer
from cocotb.utils import get_sim_time

from uvm import UVMObject, UVMComponent, UVMDebug
# from uvm.base.uvm_component import UVMComponent
from uvm.base.uvm_config_db import UVMConfigDb, UVMConfigDbOptions
from uvm.base.uvm_root import uvm_top
from uvm.base.sv import sv
from uvm.base.uvm_object_globals import UVM_HEX
from uvm.base.uvm_globals import run_test
from uvm.macros.uvm_message_defines import uvm_error

UVMDebug.no_debug()

#----------------------------------------------------------------------
# lower
#----------------------------------------------------------------------


class lower(UVMComponent):

    def __init__(self, name, parent):
        UVMComponent.__init__(self, name, parent)
        self.data = 0
        self.str = ""

    
    async def run_phase(self, phase):
        phase.raise_objection(self)
        await Timer(10)
        print("{}: {} HI".format(get_sim_time(), self.get_full_name()))
        phase.drop_objection(self)

    def get_type_name(self):
        return "lower"

    def build_phase(self, phase):
        val_arr = []
        if UVMConfigDb.get(self, "", "data", val_arr):
            self.data = val_arr[0]
        val_arr = []
        if UVMConfigDb.get(self, "", "str", val_arr):
            self.str = val_arr[0]

    def do_print(self, printer):
        printer.print_field("data", self.data, 32)
        printer.print_string("str", self.str)

#----------------------------------------------------------------------
# myunit
#----------------------------------------------------------------------


class myunit(UVMComponent):

    def __init__(self, name, parent):
        UVMComponent.__init__(self, name, parent)
        UVMConfigDbOptions.turn_on_tracing()

        self.corr_data1 = 55
        self.corr_data2 = 101
        UVMConfigDb.set(self, "l1", "str", "hi")
        UVMConfigDb.set(self, "*", "data", 0x100)
        UVMConfigDb.set(self, "l1", "data", self.corr_data1)
        UVMConfigDb.set(self, "l2", "data", self.corr_data2)

        self.l1 = lower("l1", self)
        self.l2 = lower("l2", self)
        self.l1.data = 0x30
        self.l2.data = 0x40
        self.a = [None] * 5
        for i in range(5):
            self.a[i] = i*i

    
    async def run_phase(self, phase):
        # Check config from initial block
        if self.l1.data != self.corr_data1:
            uvm_error("BADCFG", sv.sformatf("Expected l1.data = %0d, got %0d",
                self.corr_data1, self.l1.data))
        if self.l2.data != self.corr_data2:
            uvm_error("BADCFG", sv.sformatf("Expected l2.data = %0d, got %0d",
                self.corr_data2, self.l2.data))
        if self.l1.str != self.l2.str and self.l1.str != "hi":
            uvm_error("BADCFG", sv.sformatf(
                "Expected l1.str = \"hi\" and l2.str = \"hi\", got l1.str = \"%s\" and l2.str = \"%s\"",
                self.l1.str, self.l2.str))
        phase.raise_objection(self)
        await Timer(10)
        #10 $display("%0t: %s HI", $time, get_full_name());
        phase.drop_objection(self)

    def get_type_name(self):
        return "myunit"

    def do_print(self, printer):
        printer.print_array_header("a", len(self.a))
        for i in range(len(self.a)):
            printer.print_field(sv.sformatf("a[%0d]", i), self.a[i], 32, UVM_HEX, "[")
        printer.print_array_footer()

#----------------------------------------------------------------------
# mydata
#----------------------------------------------------------------------

class mydata(UVMObject):
    #`uvm_object_registry(mydata,"mydata")

    def __init__(self, name=""):
        UVMObject.__init__(self, name)

    def create(self, name="mydata"):
        d = mydata()
        d.set_name(name)
        return d

    def get_type_name(self):
        return "mydata"


#----------------------------------------------------------------------
# top
#----------------------------------------------------------------------

@cocotb.test()
async def test_module_top(dut):

    mu = myunit("mu", None)
    bar = mydata()

    uvm_top.finish_on_completion = 0
    UVMConfigDb.set(None, "mu.*", "data", 101)
    UVMConfigDb.set(None, "mu.*", "str", "hi")
    UVMConfigDb.set(None, "mu.l1", "data", 55)
    UVMConfigDb.set(None, "mu.*", "obj", bar)
    mu.print_obj()
    await run_test()
    mu.print_obj()



#
#-----------------------------------------------------------------------------
#   Copyright 2007-2011 Mentor Graphics Corporation
#   Copyright 2007-2010 Cadence Design Systems, Inc.
#   Copyright 2010 Synopsys, Inc.
#   Copyright 2019 Tuomas Poikela (tpoikela)
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#-----------------------------------------------------------------------------

from ..base.uvm_component import UVMComponent


class UVMMonitor(UVMComponent):
    """
    This class should be used as the base class for user-defined monitors.

    Deriving from uvm_monitor allows you to distinguish monitors from generic
    component types inheriting from uvm_component.  Such monitors will
    automatically inherit features that may be added to uvm_monitor in the future.
    """

    def __init__(self, name, parent):
        """
        Creates and initializes an instance of this class using the normal
        constructor arguments for `UVMComponent`: ~name~ is the name of the
        instance, and ~parent~ is the handle to the hierarchical parent, if any.
        """
        UVMComponent.__init__(self, name, parent)

    type_name = "uvm_monitor"

    def get_type_name(self):
        return UVMMonitor.type_name




from ..base.uvm_registry import UVMComponentRegistry, UVMObjectRegistry, uvm_report_warning
from ..base.sv import sv

from ..base.uvm_object_globals import (UVM_CHECK_FIELDS, UVM_COMPARE, UVM_COPY, UVM_DEFAULT,
                                       UVM_LOW, UVM_NOCOMPARE, UVM_NOCOPY, UVM_NONE, UVM_NOPACK,
                                       UVM_NOPRINT, UVM_NORECORD, UVM_PACK, UVM_PRINT, UVM_RADIX,
                                       UVM_READONLY, UVM_RECORD, UVM_REFERENCE, UVM_SETINT,
                                       UVM_SETOBJ, UVM_SETSTR, UVM_UNPACK)
from ..base.uvm_globals import uvm_report_info
from ..base.uvm_globals import uvm_is_match


def uvm_object_utils(T):
    Ts = T.__name__
    m_uvm_object_registry_internal(T,Ts)
    m_uvm_object_create_func(T, Ts)
    m_uvm_get_type_name_func(T, Ts)
    m_uvm_object_repr_func(T, Ts)
    return T


def uvm_component_utils(T):
    #if not __name__ in T:
    #    raise Exception("No __name__ found in {}".format(T))
    Ts = T.__name__
    m_uvm_component_registry_internal(T, Ts)
    m_uvm_get_type_name_func(T, Ts)
    return T


def m_uvm_get_type_name_func(T, Ts):
    setattr(T, 'type_name', Ts)

    def get_type_name(self):
        return Ts
    setattr(T, 'get_type_name', get_type_name)


def m_uvm_component_registry_internal(T,S):
    setattr(T, 'type_id', UVMComponentRegistry(T, S))

    def type_id():
        return T.type_id

    @classmethod
    def get_type(T):
        return T.type_id.get()
    setattr(T, 'get_type', get_type)

    def get_object_type(self):
        return T.type_id.get()
    setattr(T, 'get_object_type', get_object_type)


def m_uvm_object_registry_internal(T,S):
    setattr(T, 'type_id', UVMObjectRegistry(T, S))

    def type_id():
        return T.type_id

    @classmethod
    def get_type(T):
        return T.type_id.get()
    setattr(T, 'get_type', get_type)

    def get_object_type(self):
        return T.type_id.get()
    setattr(T, 'get_object_type', get_object_type)


def m_uvm_object_create_func(T,S):
    # pass
    def create(self, name=""):
        return T(name)
    setattr(T, 'create', create)


def m_uvm_object_repr_func(T,S):
    def __repr__(self):
        return T.convert2string(self)
    if hasattr(T, 'convert2string'):
        setattr(T, '__repr__', __repr__)


# Needed to make uvm_field_* macros work with similar args as in SV,
# stores the class object of last uvm_object_utils_begin(...)
__CURR_OBJ = None


def uvm_component_utils_begin(T):
    global __CURR_OBJ
    __CURR_OBJ = T
    uvm_component_utils(T)
    uvm_field_utils_start(T)

def uvm_component_utils_end(T):
    global __CURR_OBJ
    if __CURR_OBJ is not T:
        raise Exception('Expected: ' + str(__CURR_OBJ) + ' Got: ' + str(T))
    else:
        __CURR_OBJ = None
    uvm_field_utils_end(T)


def uvm_object_utils_begin(T):
    global __CURR_OBJ
    __CURR_OBJ = T
    uvm_object_utils(T)
    uvm_field_utils_start(T)


def uvm_object_utils_end(T):
    global __CURR_OBJ
    if __CURR_OBJ is not T:
        raise Exception('Expected: ' + str(__CURR_OBJ) + ' Got: ' + str(T))
    else:
        __CURR_OBJ = None
    uvm_field_utils_end(T)


def uvm_field_utils_start(T):
    # Create static member containers for var names and masks
    if not hasattr(T, "_m_uvm_field_names"):
        setattr(T, "_m_uvm_field_names", [])
    if not hasattr(T, "_m_uvm_field_masks"):
        setattr(T, "_m_uvm_field_masks", {})

    def _m_uvm_field_automation(self, rhs, what__, str__):
        from ..base.uvm_object import UVMObject
        bases = T.__bases__
        for Base in bases:
            if hasattr(Base, "_m_uvm_field_automation"):
                Base._m_uvm_field_automation(self, rhs, what__, str__)
        vals = T._m_uvm_field_names
        masks = T._m_uvm_field_masks
        _current_scopes = []  # uvm_object[$]

        T_cont = T._m_uvm_status_container
        T_cont_obj = UVMObject._m_uvm_status_container

        if what__ in [UVM_SETINT, UVM_SETSTR, UVM_SETOBJ]:
            if T_cont.m_do_cycle_check(self):
                return
            else:
                _current_scopes = T_cont.m_uvm_cycle_scopes
        # This part does the actual work
        if what__ == UVM_COPY:
            for v in vals:
                mask_v = masks[v]
                if not(mask_v & UVM_NOCOPY) and (mask_v & UVM_COPY != 0):
                    v_attr = getattr(rhs, v)
                    if hasattr(v_attr, "clone"):
                        setattr(self, v, v_attr.clone())
                    else:
                        setattr(self, v, v_attr)
        elif what__ == UVM_COMPARE:
            for v in vals:
                mask_v = masks[v]
                if not(mask_v & UVM_NOCOMPARE) and (mask_v & UVM_COMPARE != 0):
                    v_attr_rhs = getattr(rhs, v)
                    v_attr_self = getattr(self, v)
                    # Try simple equality first
                    if v_attr_rhs != v_attr_self:
                        if isinstance(v_attr_self, int):
                            T_cont.comparer.compare_field(v,
                                v_attr_self, v_attr_rhs, 0)
                        elif isinstance(v_attr_self, str):
                            T_cont.comparer.compare_string(v,
                                v_attr_self, v_attr_rhs, 0)
                        elif hasattr(v_attr_self, "compare"):
                            T_cont.comparer.compare_object(v,
                                v_attr_self, v_attr_rhs)

                        if (T_cont.comparer.result and
                                T_cont.comparer.show_max <= T_cont.comparer.result):
                            return
        elif what__ == UVM_PRINT:
            for v in vals:
                mask_v = masks[v]
                if not(mask_v & UVM_NOPRINT) and (mask_v & UVM_PRINT != 0):
                    v_attr_self = getattr(self, v)
                    if v_attr_self is None:
                        continue

                    if isinstance(v_attr_self, int):
                        T_cont.printer.print_field(v, v_attr_self,
                            sv.bits(v_attr_self), (what__ & UVM_RADIX))
                    elif isinstance(v_attr_self, UVMObject):
                        if mask_v & UVM_REFERENCE:
                            T_cont.printer.print_object_header(v, v_attr_self)
                        else:
                            T_cont.printer.print_object(v, v_attr_self)
                    elif isinstance(v_attr_self, str):
                        T_cont.printer.print_string(v, v_attr_self)
                    else:
                        raise Exception(
                            "Print not implemented yet with field macros. val: " + str(v_attr_self))
        elif what__ == UVM_SETINT:
            for v in vals:
                mask_v = masks[v]
                matched = False
                T_cont.scope.set_arg(v)
                matched = uvm_is_match(str__, T_cont.scope.get())
                if matched:
                    if mask_v & UVM_READONLY:
                        uvm_report_warning("RDONLY", sv.sformatf("Readonly argument match %s is ignored",
                         T_cont.get_full_scope_arg()), UVM_NONE)
                    else:
                        if T_cont.print_matches:
                            uvm_report_info("STRMTC", "set_int()" + ": Matched string "
                                + str__ + " to field " + T_cont.get_full_scope_arg(), UVM_LOW)
                        val = UVMObject._m_uvm_status_container.bitstream
                        setattr(self, v, val)
                        T_cont.status = 1
                T_cont.scope.unset_arg(v)
        elif what__ == UVM_SETSTR:
            for v in vals:
                mask_v = masks[v]
                T_cont.scope.set_arg(v)
                matched = uvm_is_match(str__, T_cont.scope.get())
                if matched:
                    if mask_v & UVM_READONLY:
                        uvm_report_warning("RDONLY", sv.sformatf("Readonly argument match %s is ignored",
                         T_cont.get_full_scope_arg()), UVM_NONE)
                    else:
                        if T_cont.print_matches:
                            uvm_report_info("STRMTC", "set_int()" + ": Matched string "
                                + str__ + " to field " + T_cont.get_full_scope_arg(), UVM_LOW)
                        val = UVMObject._m_uvm_status_container.stringv
                        setattr(self, v, val)
                        T_cont.status = 1
        elif what__ == UVM_SETOBJ:
            for v in vals:
                mask_v = masks[v]
                ARG = getattr(self, v)
                T_cont.scope.set_arg(v)
                if uvm_is_match(str__, T_cont.scope.get()):
                    if mask_v & UVM_READONLY:
                        uvm_report_warning("RDONLY", sv.sformatf("Readonly argument match %s is ignored",
                            T_cont.get_full_scope_arg()), UVM_NONE)
                    else:
                        if T_cont.print_matches:
                            uvm_report_info("STRMTC", "set_object()" + ": Matched string "
                                + str__ + " to field " + T_cont.get_full_scope_arg(), UVM_LOW)
                        #if sv.cast(ARG, T_cont_obj.object):
                        if T_cont_obj.object is None:
                            uvm_report_warning("UVM_SETOBJ NULL", sv.sformatf("Tried to set null obj to %s",
                                T_cont.get_full_scope_arg()), UVM_NONE)
                        setattr(self, v, T_cont_obj.object)
                        T_cont_obj.status = True
                elif ARG is not None and (mask_v & UVM_READONLY == 0):
                    cnt = 0
                    # Only traverse if there is a possible match.
                    while cnt < len(str__):
                        if (str__[cnt] == "." or str__[cnt] == "*"):
                            break
                        cnt += 1
                    if cnt != len(str__):
                        T_cont.scope.down(v)
                        ARG._m_uvm_field_automation(None, UVM_SETOBJ, str__)
                        T_cont.scope.up()

        elif what__ == UVM_CHECK_FIELDS:
            for v in vals:
                T_cont.do_field_check(v, self)
                mask_v = masks[v]
                ARG = getattr(self, v)
        elif what__ == UVM_PACK:
            for v in vals:
                mask_v = masks[v]
                if not(mask_v & UVM_NOPACK):
                    val = getattr(self, v)
                    if isinstance(val, int):
                        T_cont.packer.pack_field_int(val, sv.bits(val))
                    elif isinstance(val, UVMObject):
                        T_cont.packer.pack_object(val)
                    elif isinstance(val, str):
                        T_cont.packer.pack_string(val)
                    else:
                        raise TypeError("Unsupported type " + str(type(val)) + " for field automation")
        elif what__ == UVM_UNPACK:
            for v in vals:
                mask_v = masks[v]
                if not(mask_v & UVM_NOPACK):
                    val = getattr(self, v)
                    if isinstance(val, int):
                        rhs = T_cont.packer.unpack_field_int(sv.bits(val))
                    elif isinstance(val, UVMObject):
                        T_cont.packer.unpack_object(val)
                        rhs = val
                    elif isinstance(val, str):
                        rhs = T_cont.packer.unpack_string()
                    else:
                        raise TypeError("Unsupported type " + str(type(val)) + " for field automation")
                    setattr(self, v, rhs)
        elif what__ == UVM_RECORD:
            for v in vals:
                mask_v = masks[v]
                if not(mask_v & UVM_NORECORD):
                    print("Recording now attr " + v)
                    val = getattr(self, v)
                    if isinstance(val, int):
                        T_cont.recorder.record_field(v, val, 32)
                    elif isinstance(val, UVMObject):
                        T_cont.recorder.record_object(v, val)
                    elif isinstance(val, str):
                        T_cont.recorder.record_string(v, val)
                    else:
                        raise TypeError("Unsupported type " + str(type(val)) + " for field automation")

        if what__ in [UVM_SETINT, UVM_SETSTR, UVM_SETOBJ]:
            _current_scopes.pop()
            T_cont.uvm_cycle_scopes = _current_scopes

    setattr(T, "_m_uvm_field_automation", _m_uvm_field_automation)


def uvm_field_utils_end(T):
    pass


def uvm_field_val(name, mask):
    if not hasattr(__CURR_OBJ, name):
        vals = getattr(__CURR_OBJ, "_m_uvm_field_names")
        masks = getattr(__CURR_OBJ, "_m_uvm_field_masks")
        vals.append(name)
        masks[name] = mask
    else:
        raise Exception('uvm_field_val(): ' + str(__CURR_OBJ) +
            ' does not have property named ' + name)


def uvm_field_int(name, mask=UVM_DEFAULT):
    uvm_field_val(name, mask)


def uvm_field_string(name, mask=UVM_DEFAULT):
    uvm_field_val(name, mask)


def uvm_field_object(name, mask=UVM_DEFAULT):
    uvm_field_val(name, mask)


def uvm_field_aa(name, mask=UVM_DEFAULT):
    uvm_field_val(name, mask)


def uvm_field_aa_string_string(name, mask=UVM_DEFAULT):
    uvm_field_aa(name, mask)




import cocotb
from cocotb.triggers import Timer, Join
from cocotb.result import TestFailure
import random

from uvm.base.uvm_mailbox import UVMMailbox

from uvm.comps.uvm_test import UVMTest
from uvm.base.uvm_globals import run_test
from uvm.base.uvm_debug import *
from uvm.base.uvm_component import UVMComponent
from uvm.base.uvm_phase import *
from uvm.base.uvm_root import *
from uvm.macros.uvm_object_defines import *

class MyTest(UVMTest):
    def __init__(self, name, parent):
        UVMTest.__init__(self, name, parent)

    def build_phase(self, phase):
        UVMTest.build_phase(self, phase)
        print("Called build_phase() with", str(phase))

    def connect_phase(self, phase):
        UVMTest.connect_phase(self, phase)
        print("Called connect_phase() with", str(phase))

uvm_component_utils(MyTest)

@cocotb.test(skip=False)
async def mailbox_test(dut):
    npackets = 10
    fifo = UVMMailbox()
    #yield write_packets(fifo, 10)
    write_proc = cocotb.fork(write_packets(fifo, npackets))
    read_proc = cocotb.fork(read_packets(fifo, npackets))
    print("Waiting to join() the forked procs")
    #yield Timer(10)
    await write_proc.join(), read_proc.join()
    await Timer(10000)

class TestParentComp(UVMComponent):
    def __init__(self, name, parent):
        UVMComponent.__init__(self, name, parent)

    def build_phase(self, phase):
        UVMComponent.build_phase(self, phase)
        child1 = UVMComponent('child1', self)

class CompPhaseTest(UVMTest):
    def __init__(self, name, parent):
        UVMTest.__init__(self, name, parent)
        self.m_done_event = Event()
        self.all_done = False

    def final_phase(self, phase):
        self.all_done = True
        self.m_done_event.set()

# Runs the test without using uvm_top and the full framework
@cocotb.test(skip=False)
async def uvm_component_test(dut):
    #comp = UVMComponent('top', None)
    comp = TestParentComp('top', None)
    caller = UVMPhaseCaller(comp)
    caller.clear()
    caller.call_topdown_phase(comp, 'build_phase')
    caller.expect_call_order(['top', 'child1'])
    caller.clear()
    caller.call_bottomup_phase(comp, 'connect_phase')
    caller.expect_call_order(['child1', 'top'])
    #yield comp.run_phase({})
    await caller.call_topdown_task_phase(comp, 'run_phase')
    await Timer(100)

@cocotb.test(skip=False)
async def uvm_basic_top_test(dut):
    """ Test for checking that UVM is up and running """
    test = UVMTest('my test', None)
    test.dut = dut
    await run_test('MyTest')

@cocotb.test(skip=True)
async def uvm_phase_test(dut):
    uvm_root = UVMRoot.get()
    test_comp = CompPhaseTest('comp_phase_test', uvm_root)
    clp = UVMCmdlineProcessor.get_inst()
    print("tool=" + clp.get_tool_name() + " version=" + clp.get_tool_version())
    await run_test()
    await test_comp.m_done_event.wait()
    if not test_comp.all_done:
        raise Exception('Error in comp_phase_test')

# HELPERS


async def write_packets(fifo, n):
    nwritten = 0
    while nwritten < n:
        await fifo.put(n + 100)
        uvm_debug('', 'write_packets', "Wrote packet {} into FIFO".format(nwritten))
        nwritten += 1
        rand_delay = random.randint(1, 15)
        await Timer(rand_delay)
    uvm_debug('', 'write_packets', "Write completed".format(n))


async def read_packets(fifo, n):
    nread = 0
    while nread < n:
        q = []
        await fifo.get(q)
        print("Read packet {} into FIFO".format(nread))
        nread += 1
        if len(q) == 1:
            if q[0] != (n + 100):
                print("Error in n{}: Got: {}, Exp: {}".format(nread, q[0], n + 100))
        else:
            raise Exception('q.len must be 1 after get()')
        rand_delay = random.randint(1, 15)
        await Timer(rand_delay)
    print("read_packets completed")

class UVMPhaseCaller:
    """ Starts from a top comp, and calls given phases of all comps"""

    def __init__(self, top_comp, phases = []):
        self.comp = top_comp
        self.top_phase = UVMPhase('top')
        self.call_order = []

    def clear(self):
        self.call_order = []

    def call_topdown_phase(self, comp, name):
        if hasattr(comp, name):
            print("topdown OK for {}, {}".format(comp.get_name(), name))
            func = getattr(comp, name)
            func(self.top_phase)
            self.call_order.append(comp.get_name())
        else:
            raise Expection('No phase func {} in comp {}'.format(name,
                comp.get_name()))
        children = []
        comp.get_children(children)
        for c in children:
            self.call_topdown_phase(c, name)

    def call_bottomup_phase(self, comp, name):
        children = []
        comp.get_children(children)
        if len(children) == 0:
            if hasattr(comp, name):
                func = getattr(comp, name)
                func(self.top_phase)
                self.call_order.append(comp.get_name())
            else:
                raise Expection('No phase func {} in comp {}'.format(name,
                    comp.get_name()))
        else:
            for c in children:
                self.call_bottomup_phase(c, name)
            func = getattr(comp, name)
            func(self.top_phase)
            self.call_order.append(comp.get_name())
    
    
    async def call_topdown_task_phase(self, comp, name):
        #func = getattr(comp, name)
        #yield func(self.top_phase)
        children = []
        comp.get_children(children)
        if hasattr(comp, name):
             func = getattr(comp, name)
             await func(self.top_phase)
        else:
            raise Expection('No phase func {} in comp {}'.format(name,
                comp.get_name()))
        for c in children:
             await self.call_topdown_task_phase(c, name)

    def expect_call_order(self, exp_order):
        ii = 0
        if len(exp_order) != len(self.call_order):
            raise Exception('Call order len differs: {} vs {}'.format(
                len(exp_order), len(self.call_order)))
        for i in exp_order:
            if i != self.call_order[ii]:
                raise Exception('Call order error: Got {} Exp: {}'.format(
                    self.call_order[ii], i))
            ii += 1




#
#------------------------------------------------------------------------------
#   Copyright 2007-2011 Mentor Graphics Corporation
#   Copyright 2007-2010 Cadence Design Systems, Inc.
#   Copyright 2010 Synopsys, Inc.
#   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#------------------------------------------------------------------------------
"""
Title: TLM Export Classes

The following classes define the TLM export classes.

CLASS: UVM*Export

The unidirectional UVM*Export is a port that ~forwards~ or ~promotes~
an interface implementation from a child component to its parent.
An export can be connected to any compatible child export or imp port.
It must ultimately be connected to at least one implementation
of its associated interface.

The interface type represented by the asterisk is any of the following

  blocking_put
  nonblocking_put
  put

  blocking_get
  nonblocking_get
  get

  blocking_peek
  nonblocking_peek
  peek

  blocking_get_peek
  nonblocking_get_peek
  get_peek

Exports are connected to interface implementations directly via
UVM*Imp ports or indirectly via other UVM*Export exports.



Function: __init__

The ~name~ and ~parent~ are the standard `UVMComponent` constructor arguments.
The ~min_size~ and ~max_size~ specify the minimum and maximum number of
interfaces that must have been supplied to this port by the end of elaboration.

.. code-block:: python
    def __init__(self, name, parent, min_size=1, max_size=1):

"""

from ..base.uvm_port_base import UVMPortBase
from .uvm_tlm_imps import (UVM_BLOCKING_GET_IMP, UVM_BLOCKING_GET_PEEK_IMP, UVM_BLOCKING_PEEK_IMP,
   UVM_BLOCKING_PUT_IMP, UVM_BLOCKING_TRANSPORT_IMP, UVM_EXPORT_COMMON,
   UVM_GET_IMP, UVM_GET_PEEK_IMP, UVM_NONBLOCKING_GET_IMP,
   UVM_NONBLOCKING_GET_PEEK_IMP, UVM_NONBLOCKING_PEEK_IMP,
   UVM_NONBLOCKING_PUT_IMP, UVM_NONBLOCKING_TRANSPORT_IMP, UVM_PEEK_IMP,
   UVM_PUT_IMP, UVM_TRANSPORT_IMP)
from ..macros.uvm_tlm_defines import (UVM_TLM_BLOCKING_GET_MASK, UVM_TLM_BLOCKING_GET_PEEK_MASK,
  UVM_TLM_BLOCKING_MASTER_MASK, UVM_TLM_BLOCKING_PEEK_MASK,
  UVM_TLM_BLOCKING_PUT_MASK, UVM_TLM_BLOCKING_SLAVE_MASK,
  UVM_TLM_BLOCKING_TRANSPORT_MASK, UVM_TLM_GET_MASK,
  UVM_TLM_GET_PEEK_MASK, UVM_TLM_MASTER_MASK,
  UVM_TLM_NONBLOCKING_GET_MASK,
  UVM_TLM_NONBLOCKING_GET_PEEK_MASK,
  UVM_TLM_NONBLOCKING_MASTER_MASK,
  UVM_TLM_NONBLOCKING_PEEK_MASK, UVM_TLM_NONBLOCKING_PUT_MASK,
  UVM_TLM_NONBLOCKING_SLAVE_MASK,
  UVM_TLM_NONBLOCKING_TRANSPORT_MASK, UVM_TLM_PEEK_MASK,
  UVM_TLM_PUT_MASK, UVM_TLM_SLAVE_MASK, UVM_TLM_TRANSPORT_MASK)


#class uvm_blocking_put_export #(type T=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_EXPORT_COMMON(`UVM_TLM_BLOCKING_PUT_MASK,"uvm_blocking_put_export")
#  `UVM_BLOCKING_PUT_IMP ('m_if', T, t)
class UVMBlockingPutExport(UVMPortBase):
    pass
UVMBlockingPutExport = UVM_EXPORT_COMMON(UVMBlockingPutExport,  # type: ignore
    UVM_TLM_BLOCKING_PUT_MASK,"uvm_blocking_put_export")
UVM_BLOCKING_PUT_IMP('m_if', UVMBlockingPutExport)

#class uvm_nonblocking_put_export #(type T=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_EXPORT_COMMON(`UVM_TLM_NONBLOCKING_PUT_MASK,"uvm_nonblocking_put_export")
#  `UVM_NONBLOCKING_PUT_IMP ('m_if', T, t)
class uvm_nonblocking_put_export(UVMPortBase):
    pass
uvm_nonblocking_put_export = UVM_EXPORT_COMMON(uvm_nonblocking_put_export,  # type: ignore
    UVM_TLM_NONBLOCKING_PUT_MASK,"uvm_nonblocking_put_export")
UVM_NONBLOCKING_PUT_IMP('m_if', uvm_nonblocking_put_export)
UVMNonBlockingPutExport = uvm_nonblocking_put_export

#class uvm_put_export #(type T=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_EXPORT_COMMON(`UVM_TLM_PUT_MASK,"uvm_put_export")
#  `UVM_PUT_IMP ('m_if', T, t)
class uvm_put_export(UVMPortBase):
    pass
uvm_put_export = UVM_EXPORT_COMMON(uvm_put_export, UVM_TLM_PUT_MASK,  # type: ignore
    "uvm_put_export")
UVM_PUT_IMP('m_if', uvm_put_export)
UVMPutExport = uvm_put_export

#class uvm_blocking_get_export #(type T=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_EXPORT_COMMON(`UVM_TLM_BLOCKING_GET_MASK,"uvm_blocking_get_export")
#  `UVM_BLOCKING_GET_IMP ('m_if', T, t)
class uvm_blocking_get_export(UVMPortBase):
    pass
uvm_blocking_get_export = UVM_EXPORT_COMMON(uvm_blocking_get_export,  # type: ignore
    UVM_TLM_BLOCKING_GET_MASK,"uvm_blocking_get_export")
UVM_BLOCKING_GET_IMP('m_if', uvm_blocking_get_export)
UVMBlockingGetExport = uvm_blocking_get_export

#class uvm_nonblocking_get_export #(type T=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_EXPORT_COMMON(`UVM_TLM_NONBLOCKING_GET_MASK,"uvm_nonblocking_get_export")
#  `UVM_NONBLOCKING_GET_IMP ('m_if', T, t)
class uvm_nonblocking_get_export(UVMPortBase):
    pass
uvm_nonblocking_get_export = UVM_EXPORT_COMMON(uvm_nonblocking_get_export,  # type: ignore
    UVM_TLM_NONBLOCKING_GET_MASK,"uvm_nonblocking_get_export")
UVM_NONBLOCKING_GET_IMP('m_if', uvm_nonblocking_get_export)
UVMNonBlockingGetExport = uvm_nonblocking_get_export

#class uvm_get_export #(type T=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_EXPORT_COMMON(`UVM_TLM_GET_MASK,"uvm_get_export")
#  `UVM_GET_IMP ('m_if', T, t)
class uvm_get_export(UVMPortBase):
    pass
uvm_get_export = UVM_EXPORT_COMMON(uvm_get_export, UVM_TLM_GET_MASK,  # type: ignore
    "uvm_get_export")
UVM_GET_IMP('m_if', uvm_get_export)
UVMGetExport = uvm_get_export

#class uvm_blocking_peek_export #(type T=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_EXPORT_COMMON(`UVM_TLM_BLOCKING_PEEK_MASK,"uvm_blocking_peek_export")
#  `UVM_BLOCKING_PEEK_IMP ('m_if', T, t)
class uvm_blocking_peek_export(UVMPortBase):
    pass
uvm_blocking_peek_export = UVM_EXPORT_COMMON(uvm_blocking_peek_export,  # type: ignore
    UVM_TLM_BLOCKING_PEEK_MASK,"uvm_blocking_peek_export")
UVM_BLOCKING_PEEK_IMP('m_if', uvm_blocking_peek_export)
UVMBlockingPeekExport = uvm_blocking_peek_export

#class uvm_nonblocking_peek_export #(type T=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_EXPORT_COMMON(`UVM_TLM_NONBLOCKING_PEEK_MASK,"uvm_nonblocking_peek_export")
#  `UVM_NONBLOCKING_PEEK_IMP ('m_if', T, t)
class uvm_nonblocking_peek_export(UVMPortBase):
    pass
uvm_nonblocking_peek_export = UVM_EXPORT_COMMON(uvm_nonblocking_peek_export,  # type: ignore
    UVM_TLM_NONBLOCKING_PEEK_MASK,"uvm_nonblocking_peek_export")
UVM_NONBLOCKING_PEEK_IMP('m_if', uvm_nonblocking_peek_export)
UVMNonBlockingPeekExport = uvm_nonblocking_peek_export

#class uvm_peek_export #(type T=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_EXPORT_COMMON(`UVM_TLM_PEEK_MASK,"uvm_peek_export")
#  `UVM_PEEK_IMP ('m_if', T, t)
class uvm_peek_export(UVMPortBase):
    pass
uvm_peek_export = UVM_EXPORT_COMMON(uvm_peek_export, UVM_TLM_PEEK_MASK,  # type: ignore
    "uvm_peek_export")
UVM_PEEK_IMP ('m_if', uvm_peek_export)
UVMPeekExport = uvm_peek_export

#class uvm_blocking_get_peek_export #(type T=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_EXPORT_COMMON(`UVM_TLM_BLOCKING_GET_PEEK_MASK,"uvm_blocking_get_peek_export")
#  `UVM_BLOCKING_GET_PEEK_IMP ('m_if', T, t)
class uvm_blocking_get_peek_export(UVMPortBase):
    pass
uvm_blocking_get_peek_export = UVM_EXPORT_COMMON(uvm_blocking_get_peek_export, # type: ignore
    UVM_TLM_BLOCKING_GET_PEEK_MASK,"uvm_blocking_get_peek_export")
UVM_BLOCKING_GET_PEEK_IMP('m_if', uvm_blocking_get_peek_export)
UVMBlockingGetPeekExport = uvm_blocking_get_peek_export

#class uvm_nonblocking_get_peek_export #(type T=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_EXPORT_COMMON(`UVM_TLM_NONBLOCKING_GET_PEEK_MASK,"uvm_nonblocking_get_peek_export")
#  `UVM_NONBLOCKING_GET_PEEK_IMP ('m_if', T, t)
class uvm_nonblocking_get_peek_export(UVMPortBase):
    pass
uvm_nonblocking_get_peek_export = UVM_EXPORT_COMMON(uvm_nonblocking_get_peek_export,  # type: ignore
    UVM_TLM_NONBLOCKING_GET_PEEK_MASK,"uvm_nonblocking_get_peek_export")
UVM_NONBLOCKING_GET_PEEK_IMP ('m_if', uvm_nonblocking_get_peek_export)
UVMNonBlockingGetPeekExport = uvm_nonblocking_get_peek_export

#class uvm_get_peek_export #(type T=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_EXPORT_COMMON(`UVM_TLM_GET_PEEK_MASK,"uvm_get_peek_export")
#  `UVM_GET_PEEK_IMP ('m_if', T, t)
class uvm_get_peek_export(UVMPortBase):
    pass
uvm_get_peek_export = UVM_EXPORT_COMMON(uvm_get_peek_export,  # type: ignore
    UVM_TLM_GET_PEEK_MASK,"uvm_get_peek_export")
UVM_GET_PEEK_IMP('m_if', uvm_get_peek_export)
UVMGetPeekExport = uvm_get_peek_export

#------------------------------------------------------------------------------
#
# CLASS: uvm_*_export #(REQ,RSP)
#
# The bidirectional uvm_*_export is a port that ~forwards~ or ~promotes~
# an interface implementation from a child component to its parent.
# An export can be connected to any compatible child export or imp port.
# It must ultimately be connected to at least one implementation
# of its associated interface.
#
# The interface type represented by the asterisk is any of the following
#
#|  blocking_transport
#|  nonblocking_transport
#|  transport
#|
#|  blocking_master
#|  nonblocking_master
#|  master
#|
#|  blocking_slave
#|  nonblocking_slave
#|  slave
#
# Type parameters
#
# REQ - The type of request transaction to be communicated by the export
#
# RSP - The type of response transaction to be communicated by the export
#
# Exports are connected to interface implementations directly via
# <uvm_*_imp #(REQ, RSP, IMP, REQ_IMP, RSP_IMP)> ports or indirectly via other
# <uvm_*_export #(REQ,RSP)> exports.
#
#------------------------------------------------------------------------------
#
# Function: new
#
# The ~name~ and ~parent~ are the standard <uvm_component> constructor arguments.
# The ~min_size~ and ~max_size~ specify the minimum and maximum number of
# interfaces that must have been supplied to this port by the end of elaboration.
#
#|  function new (string name,
#|                uvm_component parent,
#|                int min_size=1,
#|                int max_size=1)

#class uvm_blocking_master_export #(type REQ=int, type RSP=REQ)
#  extends uvm_port_base #(uvm_tlm_if_base #(REQ, RSP));
#  `UVM_EXPORT_COMMON(`UVM_TLM_BLOCKING_MASTER_MASK,"uvm_blocking_master_export")
#  `UVM_BLOCKING_PUT_IMP ('m_if', REQ, t)
#  `UVM_BLOCKING_GET_PEEK_IMP ('m_if', RSP, t)
class uvm_blocking_master_export:
    pass
uvm_blocking_master_export = UVM_EXPORT_COMMON(uvm_blocking_master_export,  # type: ignore
    UVM_TLM_BLOCKING_MASTER_MASK,"uvm_blocking_master_export")
UVM_BLOCKING_PUT_IMP('m_if', uvm_blocking_master_export)
UVM_BLOCKING_GET_PEEK_IMP('m_if', uvm_blocking_master_export)
UVMBlockingMasterExport = uvm_blocking_master_export

#class uvm_nonblocking_master_export #(type REQ=int, type RSP=REQ)
#  extends uvm_port_base #(uvm_tlm_if_base #(REQ, RSP));
#  `UVM_EXPORT_COMMON(`UVM_TLM_NONBLOCKING_MASTER_MASK,"uvm_nonblocking_master_export")
#  `UVM_NONBLOCKING_PUT_IMP ('m_if', REQ, t)
#  `UVM_NONBLOCKING_GET_PEEK_IMP ('m_if', RSP, t)
class uvm_nonblocking_master_export:
    pass
uvm_nonblocking_master_export = UVM_EXPORT_COMMON(uvm_nonblocking_master_export,  # type: ignore
    UVM_TLM_NONBLOCKING_MASTER_MASK, "uvm_nonblocking_master_export")
UVM_NONBLOCKING_PUT_IMP('m_if', uvm_nonblocking_master_export)
UVM_NONBLOCKING_GET_PEEK_IMP('m_if', uvm_nonblocking_master_export)
UVMNonBlockingMasterExport = uvm_nonblocking_master_export

#class uvm_master_export #(type REQ=int, type RSP=REQ)
#  extends uvm_port_base #(uvm_tlm_if_base #(REQ, RSP));
#  `UVM_EXPORT_COMMON(`UVM_TLM_MASTER_MASK,"uvm_master_export")
#  `UVM_PUT_IMP ('m_if', REQ, t)
#  `UVM_GET_PEEK_IMP ('m_if', RSP, t)
class uvm_master_export:
    pass
uvm_master_export = UVM_EXPORT_COMMON(uvm_master_export,  # type: ignore
    UVM_TLM_MASTER_MASK, "uvm_master_port")
UVM_PUT_IMP('m_if', uvm_master_export)
UVM_GET_PEEK_IMP('m_if', uvm_master_export)
UVMMasterExport = uvm_master_export

#class uvm_blocking_slave_export #(type REQ=int, type RSP=REQ)
#  extends uvm_port_base #(uvm_tlm_if_base #(RSP, REQ));
#  `UVM_EXPORT_COMMON(`UVM_TLM_BLOCKING_SLAVE_MASK,"uvm_blocking_slave_export")
#  `UVM_BLOCKING_PUT_IMP ('m_if', RSP, t)
#  `UVM_BLOCKING_GET_PEEK_IMP ('m_if', REQ, t)
class uvm_blocking_slave_export:  #(type REQ=int, type RSP=REQ)
    pass
uvm_blocking_slave_export = UVM_EXPORT_COMMON(uvm_blocking_slave_export,  # type: ignore
    UVM_TLM_BLOCKING_SLAVE_MASK, "uvm_blocking_slave_export")
UVM_BLOCKING_PUT_IMP('m_if', uvm_blocking_slave_export)
UVM_BLOCKING_GET_PEEK_IMP('m_if', uvm_blocking_slave_export)

#class uvm_nonblocking_slave_export #(type REQ=int, type RSP=REQ)
#  extends uvm_port_base #(uvm_tlm_if_base #(RSP, REQ));
#  `UVM_EXPORT_COMMON(`UVM_TLM_NONBLOCKING_SLAVE_MASK,"uvm_nonblocking_slave_export")
#  `UVM_NONBLOCKING_PUT_IMP ('m_if', RSP, t)
#  `UVM_NONBLOCKING_GET_PEEK_IMP ('m_if', REQ, t)
class uvm_nonblocking_slave_export:  #(type REQ=int, type RSP=REQ)
    pass
uvm_nonblocking_slave_export = UVM_EXPORT_COMMON(uvm_nonblocking_slave_export, # type: ignore
    UVM_TLM_NONBLOCKING_SLAVE_MASK,"uvm_nonblocking_slave_export")
UVM_NONBLOCKING_PUT_IMP('m_if', uvm_nonblocking_slave_export)
UVM_NONBLOCKING_GET_PEEK_IMP('m_if', uvm_nonblocking_slave_export)
UVMNonBlockingSlaveExport = uvm_nonblocking_slave_export


#class uvm_slave_export #(type REQ=int, type RSP=REQ)
#  extends uvm_port_base #(uvm_tlm_if_base #(RSP, REQ));
#  `UVM_EXPORT_COMMON(`UVM_TLM_SLAVE_MASK,"uvm_slave_export")
#  `UVM_PUT_IMP ('m_if', RSP, t)
#  `UVM_GET_PEEK_IMP ('m_if', REQ, t)
class uvm_slave_export:  # (type REQ=int, type RSP=REQ)
    pass
uvm_slave_export = UVM_EXPORT_COMMON(uvm_slave_export,  # type: ignore
    UVM_TLM_SLAVE_MASK, "uvm_slave_export")
UVM_PUT_IMP('m_if', uvm_slave_export)
UVM_GET_PEEK_IMP('m_if', uvm_slave_export)
UVMSlaveExport = uvm_slave_export

#class uvm_blocking_transport_export #(type REQ=int, type RSP=REQ)
#  extends uvm_port_base #(uvm_tlm_if_base #(REQ, RSP));
#  `UVM_EXPORT_COMMON(`UVM_TLM_BLOCKING_TRANSPORT_MASK,"uvm_blocking_transport_export")
#  `UVM_BLOCKING_TRANSPORT_IMP ('m_if', REQ, RSP, req, rsp)
class uvm_blocking_transport_export:  #(type REQ=int, type RSP=REQ)
    pass
uvm_blocking_transport_export = UVM_EXPORT_COMMON(uvm_blocking_transport_export,  # type: ignore
    UVM_TLM_BLOCKING_TRANSPORT_MASK, "uvm_blocking_transport_export")
UVM_BLOCKING_TRANSPORT_IMP('m_if', uvm_blocking_transport_export)
UVMBlockingTransportExport = uvm_blocking_transport_export


#class uvm_nonblocking_transport_export #(type REQ=int, type RSP=REQ)
#  extends uvm_port_base #(uvm_tlm_if_base #(REQ, RSP));
#  `UVM_EXPORT_COMMON(`UVM_TLM_NONBLOCKING_TRANSPORT_MASK,"uvm_nonblocking_transport_export")
#  `UVM_NONBLOCKING_TRANSPORT_IMP ('m_if', REQ, RSP, req, rsp)
class uvm_nonblocking_transport_export:  #(type REQ=int, type RSP=REQ)
    pass
uvm_nonblocking_transport_export = UVM_EXPORT_COMMON(uvm_nonblocking_transport_export,  # type: ignore
    UVM_TLM_NONBLOCKING_TRANSPORT_MASK,"uvm_nonblocking_transport_export")
UVM_NONBLOCKING_TRANSPORT_IMP('m_if', uvm_nonblocking_transport_export)
UVMNonBlockingTransportExport = uvm_nonblocking_transport_export

#class uvm_transport_export #(type REQ=int, type RSP=REQ)
#  extends uvm_port_base #(uvm_tlm_if_base #(REQ, RSP));
#  `UVM_EXPORT_COMMON(`UVM_TLM_TRANSPORT_MASK,"uvm_transport_export")
#  `UVM_TRANSPORT_IMP ('m_if', REQ, RSP, req, rsp)
class uvm_transport_export:  #(type REQ=int, type RSP=REQ)
    pass
uvm_transport_export = UVM_EXPORT_COMMON(uvm_transport_export,  # type: ignore
    UVM_TLM_TRANSPORT_MASK,"uvm_transport_export")
UVM_TRANSPORT_IMP ('m_if', uvm_transport_export)
UVMTransportExport = uvm_transport_export



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2011 Synopsys, Inc.
#//    Copyright 2010-2011 Mentor Graphics Corporation
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//


from uvm.comps.uvm_env import UVMEnv
from uvm.macros import *

from apb.apb_agent import apb_agent
from apb.apb_rw import reg2apb_adapter
from reg_B import reg_block_B


class blk_env(UVMEnv):


    def __init__(self, name="blk_env", parent=None):
        super().__init__(name, parent)
        self.model = None  # reg_block_B
        self.apb = None  # apb_agent
        self.all_ok = False

    def build_phase(self, phase):
        super().build_phase(phase)

        if self.model is None:
            self.model = reg_block_B.type_id.create("reg_blk_B")
            self.model.build()
            self.model.lock_model()
            self.apb = apb_agent.type_id.create("apb",self)


    def connect_phase(self, phase):
        if self.model.get_parent() is None:
            reg2apb = reg2apb_adapter()
            self.model.default_map.set_sequencer(self.apb.sqr, reg2apb)
            self.model.default_map.set_auto_predict()


    def report_phase(self, phase):
        if self.apb is not None and self.apb.error is False:
            self.all_ok = True
        elif self.apb is not None and self.apb.error is True:
            uvm_error("BLK_ENV", "Found errors in APB agent")


uvm_component_utils(blk_env)
